from pathlib import Path

from molflux.metrics.suites.catalogue import register_suite

_CWD: Path = Path(__file__).parent.resolve()


def import_suites() -> None:
    """Load all available suites in the catalogue."""
    paths = (path for path in _CWD.iterdir() if path.suffix in {".yml", ".yaml"})
    for path in paths:
        register_suite(suite_path=path)



"""
Abstract Base Classes for classes implementing the SplittingStrategy protocol.
"""

import inspect
import types
from abc import ABC, abstractmethod
from collections.abc import Iterator
from typing import Any

from molflux import __version__
from molflux.splits.info import SplittingStrategyInfo
from molflux.splits.naming import camelcase_to_snakecase
from molflux.splits.typing import ArrayLike, SplitIndices, Splittable
from molflux.splits.utils import copyfunc


class SplittingStrategyBase(ABC):
    """The abstract base class for all concrete splitting strategies."""

    def __init__(self, tag: str | None = None, **kwargs: Any) -> None:
        """Initialises the strategy."""

        # build info
        info = self._info()
        info.name = camelcase_to_snakecase(type(self).__name__)
        info.tag = tag or info.name
        info.version = __version__
        info.split_description = info.split_description or self._split.__doc__ or ""

        # persist info
        self._splitting_strategy_info = info

        # Initialise a null state
        self._state: dict[str, Any] = {}

        # Update 'split' docstring with detailed kwargs description from .info()
        # need to copy method to avoid changing the docstring of every instance
        self.split = types.MethodType(copyfunc(self.split), self)  # type: ignore[method-assign]
        self.split.__func__.__doc__ += self._splitting_strategy_info.split_description  # type: ignore[operator]

        # The split signature
        self.split_signature = inspect.signature(self._split)

        # Safeguard against invalid kwargs
        self._init_signature = inspect.signature(self.__init__)  # type: ignore[misc]
        if kwargs:  # all kwargs should have been handled by now
            unknown_kwargs = list(kwargs.keys())
            raise ValueError(
                f"Unknown initialisation parameter(s): {unknown_kwargs}\n\n"
                f"Expected signature __init__{self._init_signature}.\n\n"
                f"Did you mean to pass these parameters to the strategy's .split() method instead?",
            )

    def __str__(self) -> str:
        return (
            f"SplittingStrategy(\n"
            f'\tname: "{self.name}",\n'
            f'\ttag: "{self.tag}",\n'
            f"\tsignature: self.split{self.split_signature},\n"
            f'\tdescription: """{self._splitting_strategy_info.description}""",\n'
            f'\tusage: """{self._splitting_strategy_info.split_description}"""\n'
            f"\tstate: {self.state!r}\n"
            ")"
        )

    @abstractmethod
    def _info(self) -> SplittingStrategyInfo:
        """Initialises the SplittingStrategyInfo object.

        To be implemented by subclasses.
        """

    @property
    def metadata(self) -> dict[str, Any]:
        return self._splitting_strategy_info.to_dict()

    @property
    def name(self) -> str:
        return self._splitting_strategy_info.name

    @property
    def state(self) -> dict[str, Any]:
        return self._state

    @property
    def tag(self) -> str:
        return self._splitting_strategy_info.tag

    def split(
        self,
        dataset: Splittable,
        y: ArrayLike | None = None,
        groups: ArrayLike | None = None,
        **kwargs: Any,
    ) -> Iterator[SplitIndices]:
        """
        Generate indices to split data into training, validation, and test set.
        """
        # Merge explicit keyword arguments with those stored in the state
        kwargs = {**self.state, **kwargs}

        # Safeguard against invalid kwargs
        if not all(k in self.split_signature.parameters for k in kwargs.keys()):
            unknown_kwargs = [
                k for k in kwargs.keys() if k not in self.split_signature.parameters
            ]
            raise ValueError(
                f"Unknown split parameter(s): {unknown_kwargs}\n\n"
                f"Expected signature self.split{self.split_signature}",
            )

        yield from self._split(dataset=dataset, y=y, groups=groups, **kwargs)

    @abstractmethod
    def _split(
        self,
        dataset: Splittable,
        y: ArrayLike | None = None,
        groups: ArrayLike | None = None,
        **kwargs: Any,
    ) -> Iterator[SplitIndices]:
        """The actual split callable to be implemented by subclasses."""

    def reset_state(self) -> None:
        self._state = {}

    def update_state(self, **kwargs: Any) -> None:
        self._state = {**self.state, **kwargs}



from unittest.mock import patch

import pytest
from openeye.oechem import OEMol, OEReadMolFromBytes

import datasets
from molflux.datasets import load_dataset
from molflux.datasets.builders.spice.spice import SPICE
from molflux.datasets.catalogue import list_datasets

dataset_name = "spice"
backend_name = "openeye"


@pytest.fixture()
def _fixture_import_main_class_patch():
    """Mocks out the import_main_class from datasets'.

    This function is used when generating a dataset builder in the `datasets` class.
    It is done by copying the file path of a builder script into a temporary folder
    under the huggingface cache directory, and loading the Builder class dynamically
    at runtime. This means that any test for a builder will instead be tested on a
    copy of the builder script, not showing up as covered in any coverage run.
    Instead, we redirect the builder class import to the actual class in datasets.
    """
    with patch("datasets.load.import_main_class") as mock:
        mock.return_value = SPICE
        yield


pytestmark = pytest.mark.usefixtures("_fixture_import_main_class_patch")


@pytest.fixture(scope="module")
def fixture_path_to_mock_data(fixture_path_to_this_test):
    """The path to the directory containing mock data for this builder."""
    here = fixture_path_to_this_test
    return here / "mock_data"


@pytest.fixture()
def _fixture_mocked_dataset_asset(fixture_path_to_mock_data):
    """Mocks out the DownloadManager used in 'datasets.load_dataset()'.

    The DownloadManager is used in the builder's '_split_generator()`, which
    is the function that actually downloads all the data for a given dataset.
    In this way we can avoid having to download the real datasets from AWS S3,
    and return paths to local assets.
    """
    with patch("datasets.builder.DownloadManager.download_and_extract") as mock:
        mock.return_value = str(fixture_path_to_mock_data / "data.h5")
        yield


def test_in_catalogue():
    """That the builder is registered in the catalogue."""
    catalogue = list_datasets()
    all_names = [name for names in catalogue.values() for name in names]
    assert dataset_name in all_names


@pytest.mark.usefixtures("_fixture_mocked_dataset_asset")
def test_returns_huggingface_dataset():
    """That the built dataset is a huggingface dataset."""
    dataset = load_dataset(dataset_name, backend_name, split="train")
    assert isinstance(dataset, datasets.Dataset)


@pytest.mark.usefixtures("_fixture_mocked_dataset_asset")
def test_dataset_has_correct_num_rows():
    """That the built dataset has correct num rows."""

    dataset = load_dataset(dataset_name, backend_name)

    assert len(dataset) == 10


@pytest.mark.usefixtures("_fixture_mocked_dataset_asset")
def test_dataset_is_readable_with_openeye():
    """That openeye can read the mol bytes"""

    dataset = load_dataset(dataset_name, backend_name)

    for mol_bytes in dataset["mol_bytes"]:
        mol = OEMol()
        OEReadMolFromBytes(mol, ".oeb", mol_bytes)


@pytest.mark.usefixtures("_fixture_mocked_dataset_asset")
def test_dataset_mols_have_correct_charges():
    """That the built dataset has correct num rows."""

    dataset = load_dataset(dataset_name, backend_name)
    list_charges = [
        [
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            1,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
        ],
        [
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            1,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
        ],
        [
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            1,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
        ],
        [
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            1,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
        ],
        [
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
        ],
        [
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
        ],
        [
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
        ],
        [
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
        ],
        [
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
        ],
        [
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
        ],
    ]

    for idx, mol_bytes in enumerate(dataset["mol_bytes"]):
        mol = OEMol()
        OEReadMolFromBytes(mol, ".oeb", mol_bytes)

        for jdx, atom in enumerate(mol.GetAtoms()):
            assert atom.GetFormalCharge() == list_charges[idx][jdx]



import logging
from collections.abc import Iterator
from typing import Any

import numpy as np
from openeye import oechem, oemedchem

from molflux.splits.bases import SplittingStrategyBase
from molflux.splits.info import SplittingStrategyInfo
from molflux.splits.typing import ArrayLike, SplitIndices, Splittable
from molflux.splits.utils import partition

logger = logging.getLogger(__name__)

_DESCRIPTION = """
Class for doing data splits based on the scaffold of small molecules using OpenEye.

Group  molecules  based on  the Bemis-Murcko scaffold representation, which identifies rings,
linkers, frameworks (combinations between linkers and rings) and atomic properties  such as
atom type, hibridization and bond order in a dataset of molecules. Then split the groups by
the number of molecules in each group in decreasing order.

IMPORTANT NOTE: If no core can be extracted, a warning will be raised and the SMILES itself is used as a scaffold.

References:
    .. [1] Bemis, Guy W., and Mark A. Murcko. "The properties of known drugs.
    1. Molecular frameworks." Journal of medicinal chemistry 39.15 (1996): 2887-2893.
"""


class Scaffold(SplittingStrategyBase):
    def _info(self) -> SplittingStrategyInfo:
        return SplittingStrategyInfo(
            description=_DESCRIPTION,
        )

    def _split(
        self,
        dataset: Splittable,
        y: ArrayLike | None = None,
        groups: ArrayLike | None = None,
        train_fraction: float = 0.8,
        validation_fraction: float = 0.1,
        test_fraction: float = 0.1,
        adjust_h_count: bool = True,
        r_group: bool = False,
        include_unsaturated_heterobonds: bool = True,
        **kwargs: Any,
    ) -> Iterator[SplitIndices]:
        """
        Args:
            y: The list of smiles to use for the scaffold split
            train_fraction: The proportion of the dataset to include in the train split.
            validation_fraction: The proportion of the dataset to include in the validation split.
            test_fraction: The proportion of the dataset to include in the test split.
            adjust_h_count: Adjust hydrogen count of the framework.
            r_group: Add R-groups as replacement for bonds broken during fragmentation.
            include_unsaturated_heterobonds: Include unsaturated heterobonds in frameworks (eg exocyclic carbonyls).

        Examples:
            >>> from molflux.splits import load_splitting_strategy
            >>> strategy = load_splitting_strategy('scaffold')
            >>> dataset = ['CCCC', 'CC', 'c1ccncc1']
            >>> folds = strategy.split(dataset=dataset, y=dataset)
        """
        if y is None:
            raise ValueError("""y parameter should be provided for scaffold splits.""")

        np.testing.assert_almost_equal(
            train_fraction + validation_fraction + test_fraction,
            1.0,
        )

        train_cutoff, validation_cutoff = partition(
            dataset,
            train_fraction,
            validation_fraction,
        )

        scaffold_sets = generate_scaffolds(
            y=y,
            adjust_h_count=adjust_h_count,
            r_group=r_group,
            include_unsaturated_heterobonds=include_unsaturated_heterobonds,
        )

        train_indices: list[int] = []
        validation_indices: list[int] = []
        test_indices: list[int] = []

        for scaffold_set in scaffold_sets:
            if len(train_indices) + len(scaffold_set) > train_cutoff:
                if (
                    len(train_indices) + len(validation_indices) + len(scaffold_set)
                    > validation_cutoff
                ):
                    test_indices += scaffold_set
                else:
                    validation_indices += scaffold_set
            else:
                train_indices += scaffold_set

        yield train_indices, validation_indices, test_indices


def _generate_scaffold(
    smiles: str,
    adjust_h_count: bool = True,
    r_group: bool = False,
    include_unsaturated_heterobonds: bool = True,
) -> str:
    mol = oechem.OEGraphMol()
    oechem.OESmilesToMol(mol, smiles)

    options = oemedchem.OEBemisMurckoOptions()

    #  SetRegionTyoe expects a string, rather than an Enum
    options.SetRegionType("Framework")

    if include_unsaturated_heterobonds:
        options.SetUnsaturatedHeteroBonds(True)

    absets = oemedchem.OEGetBemisMurcko(mol, options)
    # needed because absets is not an iterator, could be empty and not return anything or iterate in a for loop
    i = 0
    for abset in absets:  # noqa: B007
        i += 1

    if i > 0:
        framework = oechem.OEGraphMol()
        oechem.OESubsetMol(framework, mol, abset, adjust_h_count, r_group)
        scaffold_smiles: str = oechem.OEMolToSmiles(framework)
        return scaffold_smiles
    else:
        logger.warning(
            f"Could not find scaffold for molecule {smiles}. Will use the smiles itself as a scaffold.",
        )
        return smiles


def generate_scaffolds(
    y: ArrayLike,
    adjust_h_count: bool = True,
    r_group: bool = False,
    include_unsaturated_heterobonds: bool = True,
) -> list[list[int]]:
    """Returns all scaffolds from the SMILES string provided.

    Returns:
        List of indices of each scaffold in y.
    """

    scaffolds = {}
    for ind, smiles in enumerate(y):
        scaffold = _generate_scaffold(
            smiles=smiles,
            adjust_h_count=adjust_h_count,
            r_group=r_group,
            include_unsaturated_heterobonds=include_unsaturated_heterobonds,
        )

        if scaffold not in scaffolds:
            scaffolds[scaffold] = [ind]
        else:
            scaffolds[scaffold].append(ind)

    # Sort from largest to smallest scaffold sets
    scaffolds = {key: sorted(value) for key, value in scaffolds.items()}
    scaffold_sets = [
        scaffold_set
        for (scaffold, scaffold_set) in sorted(
            scaffolds.items(),
            key=lambda x: (len(x[1]), x[1][0]),
            reverse=True,
        )
    ]
    return scaffold_sets



from typing import Any

from molflux.features.bases import RepresentationBase
from molflux.features.info import RepresentationInfo
from molflux.features.typing import ArrayLike
from molflux.features.utils import assert_n_positional_args

_DESCRIPTION = """
A generic representation that splits iterable molfluxs into individual features.
"""


class Exploded(RepresentationBase):
    def _info(self) -> RepresentationInfo:
        return RepresentationInfo(
            description=_DESCRIPTION,
        )

    def _featurise(self, *columns: ArrayLike, **kwargs: Any) -> dict[str, list[Any]]:
        """Explodes each array-like sample into individual features.

        Args:
            samples: The data to featurise. Should consist of iterable molfluxs.

        Returns:
            A dictionary of exploded molfluxs.

        Examples:
            >>> from molflux.features import load_representation
            >>> representation = load_representation('exploded')
            >>> samples = [[1, 2, 3, 4], [10, 20, 30, 40], [100, 200, 300, 400]]
            >>> representation.featurise(samples)
            {'exploded::0': [1, 10, 100], 'exploded::1': [2, 20, 200], 'exploded::2': [3, 30, 300], 'exploded::3': [4, 40, 400]}
        """

        assert_n_positional_args(*columns, expected_size=1)
        samples = columns[0]

        return {
            f"{self.tag}::{i}": value
            for i, value in enumerate(map(list, zip(*samples, strict=False)))
        }



import os.path

import pytest

from molflux.modelzoo.errors import NotTrainedError
from molflux.modelzoo.protocols import Estimator
from molflux.modelzoo.store.io import load_from_store, save_to_store


def test_cannot_save_untrained_model_to_disk(tmp_path, fixture_mock_model_untrained):
    """That cannot save an untrained model to disk."""
    model = fixture_mock_model_untrained

    with pytest.raises(NotTrainedError):
        save_to_store(key=tmp_path, model=model)


def test_model_is_saved_to_disk(tmp_path, fixture_mock_model_trained):
    """That can save a trained model to disk"""
    model = fixture_mock_model_trained
    save_to_store(key=tmp_path, model=model)
    assert True


def test_creates_config_json_file(tmp_path, fixture_mock_model_trained):
    """That a config.json file is created on save"""
    model = fixture_mock_model_trained
    save_to_store(key=tmp_path, model=model)
    expected_config_file = tmp_path / "model_config.json"
    assert expected_config_file.is_file()


def test_creates_model_artefacts_dir(tmp_path, fixture_mock_model_trained):
    """That a model_artefacts/ directory is created on save"""
    model = fixture_mock_model_trained
    save_to_store(key=tmp_path, model=model)
    expected_artefacts_dir = tmp_path / "model_artefacts"
    assert expected_artefacts_dir.is_dir()


def test_loaded_from_disk(tmp_path, fixture_mock_model_trained):
    """That can load a model from disk"""
    model = fixture_mock_model_trained
    save_to_store(key=tmp_path, model=model)
    loaded_model = load_from_store(key=tmp_path)
    assert isinstance(loaded_model, Estimator)


def test_loaded_model_is_invariant(tmp_path, fixture_mock_model_trained):
    """That a model loaded from store encodes all the original information."""
    model = fixture_mock_model_trained
    save_to_store(key=tmp_path, model=model)
    loaded_model = load_from_store(key=tmp_path)
    assert loaded_model.name == model.name
    assert loaded_model.tag == model.tag
    assert loaded_model.config == model.config


def test_load_from_non_existent_directory_does_not_create_spurious_artefacts():
    """That attempting to load a model from an inexistent directory does not
    leave behind a partially initialised folder.
    """
    non_existent_dir = "does/not/exist"

    try:
        load_from_store(key=non_existent_dir)
    except FileNotFoundError:
        pass

    assert not os.path.exists(non_existent_dir)



from typing import Any

import pytest

import datasets
from molflux.modelzoo.catalogue import register_model
from molflux.modelzoo.info import ModelInfo
from molflux.modelzoo.load import load_model
from molflux.modelzoo.model import ModelBase, ModelConfig
from molflux.modelzoo.protocols import Model


class MockModelConfig(ModelConfig): ...


# Registers a mock model in the catalogue
@register_model(kind="pytest", name="mock_model")
class MockModel(ModelBase[MockModelConfig]):
    @property
    def _config_builder(self) -> type[MockModelConfig]:
        return MockModelConfig

    def _info(self) -> ModelInfo:
        return ModelInfo(
            model_description="A mock model",
        )

    def _train(self, train_data: datasets.Dataset, **kwargs: Any) -> None:
        self.model = True

        if "validation_data" in kwargs:
            raise ValueError("Should not call with validation_data.")

    def _predict(self, data: datasets.Dataset, **kwargs: Any) -> dict[str, Any]:
        return {self.tag: "Spooky Result!"}


@register_model(kind="pytest", name="mock_model_with_validation_data_arg")
class MockModelWithValidationDataArg(MockModel):
    def _train(
        self,
        train_data: datasets.Dataset,
        validation_data: datasets.Dataset | None = None,
        **kwargs: Any,
    ) -> None:
        self.model = True


@register_model(kind="pytest", name="mock_model_multi_data")
class MockModelMultiData(MockModel):
    _train_multi_data_enabled: bool = True

    def _train(
        self,
        train_data: datasets.Dataset,
        **kwargs: Any,
    ) -> None:
        raise NotImplementedError

    def _train_multi_data(
        self,
        train_data: dict[str | None, datasets.Dataset],
        **kwargs: Any,
    ) -> None:
        self.model = True

        if "validation_data" in kwargs:
            raise ValueError("Should not call with validation_data.")


@register_model(kind="pytest", name="mock_model_multi_data_with_validation_data_arg")
class MockModelMultiDataWithValidationDataArg(MockModel):
    _train_multi_data_enabled: bool = True

    def _train(
        self,
        train_data: datasets.Dataset,
        **kwargs: Any,
    ) -> None:
        raise NotImplementedError

    def _train_multi_data(
        self,
        train_data: dict[str | None, datasets.Dataset],
        validation_data: dict[str | None, datasets.Dataset] | None = None,
        **kwargs: Any,
    ) -> None:
        self.model = True


@pytest.fixture()
def fixture_mock_model_untrained() -> Model:
    """Returns an untrained mock model."""
    return load_model("mock_model", tag="custom_tag")


@pytest.fixture()
def fixture_mock_model_trained() -> Model:
    """Returns a trained mock model."""
    model = load_model(
        "mock_model",
        tag="custom_tag",
        x_features=["x"],
        y_features=["y"],
    )
    model.train(datasets.Dataset.from_dict({"x": [1], "y": [1]}))
    return model



import pytest
from openeye import oechem

from molflux.features.representations.openeye._utils import (
    get_sd_data,
    iterate_mols_or_confs,
    to_oemol,
    to_smiles,
)


def test_smiles_to_smiles():
    """That can convert a SMILES string to_smiles is invariant."""
    molecule = (
        "Cc1ccc2c(=O)c3cccc(CC(=O)OC4OC(C(=O)O)[C@@H](O)[C@@H](O)[C@@H]4O)c3oc2c1C"
    )
    out = to_smiles(molecule)
    assert isinstance(out, str)
    assert out == molecule


def test_oemol_to_smiles():
    """That can convert a OEMol object to SMILES."""
    expected_smiles = (
        "Cc1ccc2c(=O)c3cccc(c3oc2c1C)CC(=O)OC4[C@H]([C@@H]([C@@H](C(O4)C(=O)O)O)O)O"
    )
    oemol = oechem.OEMol()
    oechem.OESmilesToMol(oemol, expected_smiles)

    out = to_smiles(oemol)
    assert isinstance(out, str)
    assert out == expected_smiles


def test_bytes_to_smiles():
    """That can convert a bytes string to SMILES"""
    # Acetic Acid
    bytestring = b"\x0b\xce\n\x94\x13\x8f\x84CCOO\x83\x00\x01\x01\x01\x02\x02\x01\x03\x01 \x81\x00\x10\xb3\x81\x80?\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00!\x81\x00"

    out = to_smiles(bytestring)
    assert isinstance(out, str)
    assert out == "CC(=O)O"


def test_hex_to_smiles():
    """That can convert a hex string to SMILES"""
    # Acetic Acid
    hexstring = "0bce0a94138f8443434f4f8300010101020201030120810010b381803f000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000218100"

    out = to_smiles(hexstring)
    assert isinstance(out, str)
    assert out == "CC(=O)O"


def test_smiles_to_oemol():
    """That can convert a SMILES string to OEMol object."""
    molecule = "CC(=O)O"

    out = to_oemol(molecule)
    assert isinstance(out, oechem.OEMol)

    out_as_smiles = to_smiles(out)
    assert out_as_smiles == molecule


def test_oemol_to_oemol():
    """That converting a OEMol object to_oemol is invariant."""
    expected_smiles = "CC(=O)O"
    oemol = oechem.OEMol()
    oechem.OESmilesToMol(oemol, expected_smiles)

    out = to_oemol(oemol)
    assert isinstance(out, oechem.OEMol)

    out_as_smiles = to_smiles(expected_smiles)
    assert out_as_smiles == expected_smiles


def test_bytes_to_oemol():
    """That can convert a bytes string to OEMol object"""
    # Acetic Acid
    bytestring = b"\x0b\xce\n\x94\x13\x8f\x84CCOO\x83\x00\x01\x01\x01\x02\x02\x01\x03\x01 \x81\x00\x10\xb3\x81\x80?\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00!\x81\x00"

    out = to_oemol(bytestring)
    assert isinstance(out, oechem.OEMol)

    out_as_smiles = to_smiles(out)
    assert out_as_smiles == "CC(=O)O"


@pytest.mark.parametrize(
    ("sd_data", "conformer_sd_data", "dtype", "expected", "sd_tag", "conformer"),
    [
        (
            {"tag": "10"},
            {},
            str,
            "10",
            "tag",
            True,
        ),
        (
            {},
            {"tag": "10"},
            str,
            "10",
            "tag",
            True,
        ),
        (
            {"tag": "10"},
            {"tag": "20"},
            str,
            "20",
            "tag",
            True,
        ),
        (
            {"tag": "10"},
            {"tag": "20"},
            str,
            "10",
            "tag",
            False,
        ),
        (
            {"tag": "10"},
            {"tag": "20"},
            float,
            10.0,
            "tag",
            False,
        ),
        (
            {},
            {},
            float,
            Exception,
            "tag",
            False,
        ),
    ],
    ids=[
        "data-on-mol-not-conf",
        "data-on-conf-not-mol",
        "data-on-both-conf-supplied",
        "data-on-both-mol-supplied",
        "dtype-cast-float",
        "error-no-data",
    ],
)
def test_get_sd_data_multi_conf_mols(
    sd_data,
    conformer_sd_data,
    dtype,
    expected,
    sd_tag,
    conformer,
):
    """Test to check that the `get_sd_data` function works as expected with
    multi conformer molecules. Namely that if a conformer is supplied its SD data
    is retrieved (if present) falling back to the parent mol if not (and vice versa
    for parent molecule and conformer), as well as data type casting and erroring
    if no data is present."""
    mol = oechem.OEMol()
    oechem.OESmilesToMol(mol, "CC")
    for tag, value in sd_data.items():
        oechem.OESetSDData(mol, tag, str(value))
    conf = mol.GetActive()
    for tag, value in conformer_sd_data.items():
        oechem.OESetSDData(conf, tag, str(value))

    if expected is not Exception:
        value = get_sd_data(conf if conformer else mol, tag=sd_tag, dtype=dtype)
        assert value == expected
    else:
        with pytest.raises(ValueError, match=f"Unable to find SD tag {sd_tag!s} .*"):
            get_sd_data(conf if conformer else mol, tag=sd_tag, dtype=dtype)


def test_get_sd_data_graphmol():
    """Test that retrieval of SD data from graphmols works as expected."""
    mol = oechem.OEGraphMol()
    oechem.OESmilesToMol(mol, "CC")
    sd_data = {"tag": "20"}
    for tag, value in sd_data.items():
        oechem.OESetSDData(mol, tag, str(value))

    value = get_sd_data(mol, tag="tag", dtype=float)
    assert value == 20.0


@pytest.mark.parametrize(
    ("mols", "expand_conformers", "num", "expected_type"),
    [
        (
            [
                {"type": oechem.OEMol, "num_confs": 3},
                {"type": oechem.OEMol, "num_confs": 1},
            ],
            True,
            4,
            oechem.OEConfBase,
        ),
        (
            [
                {"type": oechem.OEMol, "num_confs": 3},
                {"type": oechem.OEMol, "num_confs": 1},
            ],
            False,
            2,
            oechem.OEConfBase,
        ),
        (
            [
                {"type": oechem.OEGraphMol},
                {"type": oechem.OEGraphMol},
            ],
            True,
            2,
            oechem.OEGraphMol,
        ),
    ],
    ids=["oemols-expand-each-conf", "oemols-active-conf", "graphmols"],
)
def test_iterate_mols_or_confs(mols, expand_conformers, num, expected_type):
    """Tests that iterating through `oemol` object conformers works as expected."""

    def create_mols(mol_dicts):
        for mol_dict in mol_dicts:
            m = mol_dict["type"]()
            oechem.OESmilesToMol(m, "CC")
            for _ in range(max(mol_dict.get("num_confs", 0) - 1, 0)):
                m.NewConf()
            yield m

    mols = create_mols(mols)

    count = 0
    for expanded_mol in iterate_mols_or_confs(
        mols,
        expand_conformers=expand_conformers,
    ):
        count += 1
        assert isinstance(expanded_mol, expected_type)

    assert count == num


def test_hex_to_oemol():
    """That can convert a hex string to OEMol object"""
    # Acetic Acid
    hexstring = "0bce0a94138f8443434f4f8300010101020201030120810010b381803f000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000218100"

    out = to_oemol(hexstring)
    assert isinstance(out, oechem.OEMol)

    out_as_smiles = to_smiles(out)
    assert out_as_smiles == "CC(=O)O"



from dataclasses import asdict
from typing import Any, Literal

import numpy as np
from pydantic.v1 import dataclasses

from molflux.modelzoo.info import ModelInfo
from molflux.modelzoo.models.pyod import (
    PyODClassificationMixin,
    PyODModelBase,
    PyODModelConfig,
)

try:
    from pyod.models.cblof import CBLOF
except ImportError as e:
    from molflux.modelzoo.errors import ExtrasDependencyImportError

    raise ExtrasDependencyImportError("pyod", e) from None


_DESCRIPTION = """
The CBLOF operator calculates the outlier score based on cluster-based
local outlier factor.

CBLOF takes as an input the data set and the cluster model that was
generated by a clustering algorithm. It classifies the clusters into small
clusters and large clusters using the parameters alpha and beta.
The anomaly score is then calculated based on the size of the cluster the
point belongs to as well as the distance to the nearest large cluster.

Use weighting for outlier factor based on the sizes of the clusters as
proposed in the original publication. Since this might lead to unexpected
behavior (outliers close to small clusters are not found), it is disabled
by default.Outliers scores are solely computed based on their distance to
the closest large cluster center.

By default, kMeans is used for clustering algorithm instead of
Squeezer algorithm mentioned in the original paper for multiple reasons.

See :cite:`he2003discovering` for details.
"""

_CONFIG_DESCRIPTION = """
Parameters
----------
n_clusters : int, optional (default=8)
    The number of clusters to form as well as the number of
    centroids to generate.

contamination : float in (0., 0.5), optional (default=0.1)
    The amount of contamination of the data set,
    i.e. the proportion of outliers in the data set. Used when fitting to
    define the threshold on the decision function.

clustering_estimator : Estimator, optional (default=None)
    The base clustering algorithm for performing data clustering.
    A valid clustering algorithm should be passed in. The estimator should
    have standard sklearn APIs, fit() and predict(). The estimator should
    have attributes ``labels_`` and ``cluster_centers_``.
    If ``cluster_centers_`` is not in the attributes once the model is fit,
    it is calculated as the mean of the samples in a cluster.

    If not set, CBLOF uses KMeans for scalability. See
    https://scikit-learn.org/stable/modules/generated/sklearn.cluster.KMeans.html

alpha : float in (0.5, 1), optional (default=0.9)
    Coefficient for deciding small and large clusters. The ratio
    of the number of samples in large clusters to the number of samples in
    small clusters.

beta : int or float in (1,), optional (default=5).
    Coefficient for deciding small and large clusters. For a list
    sorted clusters by size `|C1|, |C2|, ..., |Cn|, beta = |Ck|/|Ck-1|`

use_weights : bool, optional (default=False)
    If set to True, the size of clusters are used as weights in
    outlier score calculation.

check_estimator : bool, optional (default=False)
    If set to True, check whether the base estimator is consistent with
    sklearn standard.

    .. warning::
        check_estimator may throw errors with scikit-learn 0.20 above.

random_state : int, np.random.Generator or None, optional (default=None)
    If int, random_state is the seed used by the random
    number generator; If RandomState instance, random_state is the random
    number generator; If None, the random number generator is the
    RandomState instance used by `np.random`.
"""

Method = Literal["default", "fast"]


class Config:
    arbitrary_types_allowed = True
    extra = "forbid"


@dataclasses.dataclass(config=Config)
class CBLOFDetectorConfig(PyODModelConfig):
    n_clusters: int = 8
    contamination: float = 0.1
    clustering_estimator: Any = None
    alpha: float = 0.9
    beta: float = 5
    use_weights: bool = False
    check_estimator: bool = False
    random_state: int | np.random.Generator | None = None


class CBLOFDetector(PyODClassificationMixin, PyODModelBase[CBLOFDetectorConfig]):
    @property
    def config(self) -> dict[str, Any]:
        return asdict(self.model_config)

    @property
    def _config_builder(self) -> type[CBLOFDetectorConfig]:
        return CBLOFDetectorConfig

    def _info(self) -> ModelInfo:
        return ModelInfo(
            model_description=_DESCRIPTION,
            config_description=_CONFIG_DESCRIPTION,
        )

    def _instantiate_model(self) -> CBLOF:
        config = self.model_config
        return CBLOF(
            n_clusters=config.n_clusters,
            contamination=config.contamination,
            clustering_estimator=config.clustering_estimator,
            alpha=config.alpha,
            beta=config.beta,
            use_weights=config.use_weights,
            check_estimator=config.check_estimator,
            random_state=config.random_state,
        )



from typing import Any

from pydantic.v1 import dataclasses
from scipy import stats

import datasets
from molflux.modelzoo.info import ModelInfo
from molflux.modelzoo.model import (
    ClassificationMixin,
    ModelBase,
    ModelConfig,
)
from molflux.modelzoo.models.pystan.utils import StanWrapper
from molflux.modelzoo.typing import Classes, PredictionResult
from molflux.modelzoo.utils import (
    get_concatenated_array,
    pick_features,
    validate_features,
)

try:
    import stan
except ImportError as e:
    from molflux.modelzoo.errors import ExtrasDependencyImportError

    raise ExtrasDependencyImportError("pystan", e) from None


_DESCRIPTION = """
Implementation of bayesian ordinal classifier. Similar to a logistic regression model
See Williams et al 2019 Predicting Drug-Induced Liver Injury with Bayesian Machine Learning
"""

_CONFIG_DESCRIPTION = """
Parameters
----------
num_classes: int, default=3 Number of discrete ordinal classes (eg. equivalent to low, medium high)
num_chains: int, default=4 Number of MCMC chains to run. These can be run in parallel allowing faster sampling,
    and having multiple chains is also useful in diagnosing convergence.
num_warmup: int, default=1000 Number of warmup/burnin samples to run to help ensure sampling
    from stationary distribution.
num_samples: int, default=1000 Number of MCMC samples produced per chain. More samples allows for more accurate
    estimation of expectations, and will help ensure convergence, but will take longer to compute.
random_seed: Optional[int], default=None Random seed to be used during sampling for reproducibility.
sigma_prior: float, default=1.0 Scale parameter for prior on sigma.
mu_prior: float, default=2.0 Scale parameter for prior on mu.
"""


class Config:
    arbitrary_types_allowed = True


@dataclasses.dataclass(config=Config)
class OrdinalClassifierConfig(ModelConfig):
    num_classes: int = 3
    num_chains: int = 4
    num_warmup: int = 1000
    num_samples: int = 1000
    random_seed: int | None = None
    sigma_prior: float = 1.0
    mu_prior: float = 2.0

    def __post_init_post_parse__(self) -> None:
        if self.y_features and len(self.y_features) != 1:
            raise NotImplementedError(
                f"This model architecture only supports single task regression for now: got {self.y_features}",
            )


class OrdinalClassifier(
    ClassificationMixin,
    ModelBase[OrdinalClassifierConfig],
):
    @property
    def _config_builder(self) -> type[OrdinalClassifierConfig]:
        return OrdinalClassifierConfig

    def _info(self) -> ModelInfo:
        return ModelInfo(
            model_description=_DESCRIPTION,
            config_description=_CONFIG_DESCRIPTION,
        )

    def _instantiate_model(self) -> StanWrapper:
        return StanWrapper("OrdinalClassifier.stan")

    @property
    def classes(self) -> Classes:
        classes = [range(self.model_config.num_classes)]
        return {
            task: list(class_int)
            for task, class_int in zip(self.y_features, classes, strict=False)
        }

    def _train(self, train_data: datasets.Dataset, **kwargs: Any) -> Any:
        validate_features(train_data, self.y_features)

        x_data = pick_features(train_data, self.x_features)
        X = get_concatenated_array(x_data, self.x_features)

        y_data = pick_features(train_data, self.y_features)
        y = get_concatenated_array(y_data, self.y_features)
        y = y.ravel()

        N, p = X.shape
        stan_data = {
            "n_classes": self.model_config.num_classes,
            "N": N,
            "P": p,
            "X": X,
            "sigma_prior": self.model_config.sigma_prior,
            "mu_prior": self.model_config.mu_prior,
            "y": y,
        }

        self.model = self._instantiate_model()
        posterior = stan.build(
            program_code=self.model.model_code,
            data=stan_data,
            random_seed=self.model_config.random_seed,
        )
        fit = posterior.sample(
            num_chains=self.model_config.num_chains,
            num_warmup=self.model_config.num_warmup,
            num_samples=self.model_config.num_samples,
        )
        self.model.fit = fit

    def _predict(self, data: datasets.Dataset, **kwargs: Any) -> PredictionResult:
        validate_features(data, self.x_features)
        display_names = self._predict_display_names

        if not len(data):
            return {display_name: [] for display_name in display_names}

        x_data = pick_features(data, self.x_features)
        X_test = get_concatenated_array(x_data, self.x_features)
        N_pred, p = X_test.shape

        fit_df = self.model.fit.to_frame()
        cutpoint_cols = [col for col in fit_df.columns if col.startswith("cutpoints")]
        cutpoints = fit_df[cutpoint_cols].to_numpy()
        beta_cols = [col for col in fit_df.columns if col.startswith("beta")]
        beta = fit_df[beta_cols].to_numpy()
        N_samples = self.model_config.num_samples * self.model_config.num_chains
        stan_data = {
            "n_classes": self.model_config.num_classes,
            "N_samples": N_samples,
            "N_pred": N_pred,
            "X_pred": X_test,
            "P": p,
            "cutpoints": cutpoints,
            "beta": beta,
        }

        prediction_model = StanWrapper("OrdinalClassifierPredictions.stan")
        posterior = stan.build(
            program_code=prediction_model.model_code,
            data=stan_data,
            random_seed=self.model_config.random_seed,
        )
        fit = posterior.fixed_param(
            num_chains=1,
            num_samples=1,
        )
        post_pred_df = fit.to_frame()
        predictions = [
            int(
                stats.mode(
                    [
                        post_pred_df[f"ypred.{i}.{j}"][0]
                        for i in range(1, N_samples + 1)
                    ],
                    keepdims=False,
                ).mode,
            )
            for j in range(1, N_pred + 1)
        ]
        return {display_name: predictions for display_name in display_names}

    def _predict_proba(self, data: datasets.Dataset, **kwargs: Any) -> PredictionResult:
        validate_features(data, self.x_features)
        display_names = self._predict_proba_display_names

        if not len(data):
            return {display_name: [] for display_name in display_names}

        x_data = pick_features(data, self.x_features)
        X_test = get_concatenated_array(x_data, self.x_features)
        N_pred, p = X_test.shape

        fit_df = self.model.fit.to_frame()
        cutpoint_cols = [col for col in fit_df.columns if col.startswith("cutpoints")]
        cutpoints = fit_df[cutpoint_cols].to_numpy()
        beta_cols = [col for col in fit_df.columns if col.startswith("beta")]
        beta = fit_df[beta_cols].to_numpy()
        N_samples = self.model_config.num_samples * self.model_config.num_chains
        stan_data = {
            "n_classes": self.model_config.num_classes,
            "N_samples": N_samples,
            "N_pred": N_pred,
            "X_pred": X_test,
            "P": p,
            "cutpoints": cutpoints,
            "beta": beta,
        }

        prediction_model = StanWrapper("OrdinalClassifierPredictions.stan")
        posterior = stan.build(
            program_code=prediction_model.model_code,
            data=stan_data,
            random_seed=self.model_config.random_seed,
        )
        fit = posterior.fixed_param(
            num_chains=1,
            num_samples=1,
        )
        post_pred_df = fit.to_frame()

        all_predictions = [
            [post_pred_df[f"ypred.{i}.{j}"][0] for i in range(1, N_samples + 1)]
            for j in range(1, N_pred + 1)
        ]

        results = {}
        for task, display_name in zip(
            self.y_features,
            display_names,
            strict=False,
        ):
            prob_preds = [
                sum(pred == ii for pred in preds) / len(preds)
                for preds in all_predictions
                for ii in self.classes[task]
            ]
            results[display_name] = prob_preds

        # (n_tasks, n_samples, n_classes)
        return results



import pytest

from molflux.splits.load import load_splitting_strategy


@pytest.fixture(scope="module")
def fixture_mock_splitting_strategy():
    return load_splitting_strategy(name="linear_split")



import inspect
import logging
from abc import ABC, abstractmethod
from dataclasses import asdict
from typing import Any

import evaluate
import numpy as np

import datasets
from molflux.metrics.typing import ArrayLike, MetricResult

logger = logging.getLogger(__name__)


class HFMetric(evaluate.Metric, ABC):
    """A base class for metrics based on the evaluate.Metric API."""

    def __init__(
        self,
        tag: str | None = None,
        config_name: str | None = None,
        keep_in_memory: bool = False,
        seed: int | None = None,
        experiment_id: str | None = None,
        download_config: datasets.DownloadConfig | None = None,
        **kwargs: Any,
    ) -> None:
        # Initialise the evaluate.Metric
        super().__init__(
            config_name=config_name,
            keep_in_memory=keep_in_memory,
            cache_dir=None,
            num_process=1,
            process_id=0,
            seed=seed,
            experiment_id=experiment_id,
            max_concurrent_cache_files=10000,
            timeout=100,
            **kwargs,
        )

        # Download and prepare resources for the metric
        self.download_and_prepare(download_config=download_config)

        self._tag = tag or self.name

        # Initialise a null state
        self._state: dict[str, Any] = {}

        # The metric-specific compute signature
        self._signature = inspect.signature(self._score)

    def __bool__(self) -> bool:
        return True

    def __len__(self) -> int:
        length: int = super().__len__()
        return length

    def __str__(self) -> str:
        return (
            f"Metric(\n"
            f'\tname: "{self.name}",\n'
            f'\ttag: "{self.tag}",\n'
            f"\tfeatures: {self.features},\n"
            f"\tsignature: self.compute{self._signature},\n"
            f'\tdescription: """{self.description}""",\n'
            f'\tusage: """{self.inputs_description}"""\n'
            f"\tstate: {self.state!r}\n"
            ")"
        )

    @property
    def metadata(self) -> dict[str, Any]:
        return asdict(super().info)

    @abstractmethod
    def _info(self) -> evaluate.MetricInfo:
        """Initialises the MetricInfo metadata object for the given metric."""

    @property
    def name(self) -> str:
        metric_name: str = super().name
        return metric_name

    @property
    def tag(self) -> str:
        return self._tag

    @property
    def state(self) -> dict[str, Any]:
        return self._state

    def add_batch(
        self,
        *,
        predictions: ArrayLike | None = None,
        references: ArrayLike | None = None,
        **kwargs: Any,
    ) -> None:
        """Add a batch of predictions and references for the metric's stack."""
        predictions, references = self._pre_process_inputs(
            predictions=predictions,
            references=references,
        )
        super().add_batch(predictions=predictions, references=references, **kwargs)

    def compute(
        self,
        *,
        predictions: ArrayLike | None = None,
        references: ArrayLike | None = None,
        **kwargs: Any,
    ) -> MetricResult:
        """
        Computes the metric.
        Args:
            predictions:
            references:
            **kwargs: Can include a 'mask_by_references' kwarg that allows for masking
                the inputs by the valid references values

        Returns:

        """

        # Merge explicit keyword arguments with those stored in config
        if self._state is not None:
            kwargs = {**self._state, **kwargs}

        if "mask_by_references" in kwargs:
            mask_by_references = kwargs.pop("mask_by_references")
        else:
            mask_by_references = False

        # Safeguard against invalid kwargs
        if not all(k in self._signature.parameters for k in kwargs):
            unknown_kwargs = [k for k in kwargs if k not in self._signature.parameters]
            raise ValueError(
                f"Unknown compute parameter(s): {unknown_kwargs}\n\n"
                f"Expected signature self.compute{self._signature}",
            )

        if mask_by_references:
            predictions, references = self._mask_by_references(
                predictions=predictions,
                references=references,
            )

        predictions, references = self._pre_process_inputs(
            predictions=predictions,
            references=references,
        )

        # wrap huggingface's implementation
        try:
            # This loads the data from the cache into self.data,
            # and then calls ._compute()
            result: MetricResult | None = super().compute(
                predictions=predictions,
                references=references,
                **kwargs,
            )

            # On distributed systems, huggingface would return None
            if result is None:
                raise RuntimeError

        except Exception as exc:
            raise RuntimeError(f"Error computing metric {self.tag!r}") from exc

        return result

    def _compute(
        self,
        *,
        predictions: ArrayLike | None = None,
        references: ArrayLike | None = None,
        **kwargs: Any,
    ) -> MetricResult:
        """The callable invoked by .compute() to compute the metric."""

        # Do some checks common to all metrics
        if predictions is None or references is None:
            raise RuntimeError(
                f"""Could not compute metric {self.tag}: no data found in the stack.""",
            )

        return self._score(
            predictions=predictions,
            references=references,
            **kwargs,
        )

    @abstractmethod
    def _score(
        self,
        *,
        predictions: ArrayLike,
        references: ArrayLike,
        **kwargs: Any,
    ) -> MetricResult:
        """The callable invoked to compute the metric. To be implemented by subclasses."""

    def reset_state(self) -> None:
        """Resets the state."""
        self._state = {}

    def update_state(self, **kwargs: Any) -> None:
        """Pre-configures keyword arguments for compute()."""
        self._state = {**self._state, **kwargs}

    def _mask_by_references(
        self,
        predictions: ArrayLike | None = None,
        references: ArrayLike | None = None,
    ) -> tuple[ArrayLike | None, ...]:
        if (references is not None) and (predictions is not None):
            # mask if None or NaN
            mask = [not (x is None or np.isnan(x)) for x in references]
            references = [
                x for valid, x in zip(mask, references, strict=False) if valid
            ]
            predictions = [
                x for valid, x in zip(mask, predictions, strict=False) if valid
            ]

        return predictions, references

    def _pre_process_inputs(
        self,
        predictions: ArrayLike | None = None,
        references: ArrayLike | None = None,
    ) -> tuple[ArrayLike | None, ...]:
        """An optional callable used to process inputs before validation.

        For most use cases, there is no need to overwrite this function.
        Defaults to pass-through.
        """
        return predictions, references


class UncertaintyMetric(HFMetric):
    """
    A class for uncertainty metrics.
    Same as HFMetric but with a different signature (eg. for .compute()) so can take in prediction intervals
    """

    def __init__(
        self,
        tag: str | None = None,
        config_name: str | None = None,
        keep_in_memory: bool = False,
        seed: int | None = None,
        experiment_id: str | None = None,
        download_config: datasets.DownloadConfig | None = None,
        **kwargs: Any,
    ) -> None:
        # Initialise the HFMetric
        super().__init__(
            tag=tag,
            config_name=config_name,
            keep_in_memory=keep_in_memory,
            seed=seed,
            experiment_id=experiment_id,
            download_config=download_config,
            **kwargs,
        )

    def compute(
        self,
        *,
        predictions: ArrayLike | None = None,
        references: ArrayLike | None = None,
        standard_deviations: ArrayLike | None = None,
        prediction_intervals: ArrayLike | None = None,
        **kwargs: Any,
    ) -> MetricResult:
        """Computes the metric."""

        # Merge explicit keyword arguments with those stored in config
        if self._state is not None:
            kwargs = {**self._state, **kwargs}

        # Safeguard against invalid kwargs
        if not all(k in self._signature.parameters for k in kwargs):
            unknown_kwargs = [k for k in kwargs if k not in self._signature.parameters]
            raise ValueError(
                f"Unknown compute parameter(s): {unknown_kwargs}\n\n"
                f"Expected signature self.compute{self._signature}",
            )

        predictions, references = self._pre_process_inputs(
            predictions=predictions,
            references=references,
        )

        # wrap huggingface's implementation
        try:
            # This loads the data from the cache into self.data,
            # and then calls ._compute()
            result: MetricResult | None = super().compute(
                predictions=predictions,
                references=references,
                standard_deviations=standard_deviations,
                prediction_intervals=prediction_intervals,
                **kwargs,
            )

            # On distributed systems, huggingface would return None
            if result is None:
                raise RuntimeError

        except Exception as exc:
            raise RuntimeError(f"Error computing metric {self.tag!r}") from exc

        return result

    def _compute(
        self,
        *,
        predictions: ArrayLike | None = None,
        references: ArrayLike | None = None,
        standard_deviations: ArrayLike | None = None,
        prediction_intervals: ArrayLike | None = None,
        **kwargs: Any,
    ) -> MetricResult:
        """The callable invoked by .compute() to compute the metric."""

        # Do some checks common to all metrics
        if predictions is None or references is None:
            raise RuntimeError(
                f"""Could not compute metric {self.tag}: no data found in the stack.""",
            )

        return self._score(
            predictions=predictions,
            references=references,
            standard_deviations=standard_deviations,
            prediction_intervals=prediction_intervals,
            **kwargs,
        )

    @abstractmethod
    def _score(
        self,
        *,
        predictions: ArrayLike,
        references: ArrayLike,
        standard_deviations: ArrayLike | None = None,
        prediction_intervals: ArrayLike | None = None,
        **kwargs: Any,
    ) -> MetricResult:
        """The callable invoked to compute the metric. To be implemented by subclasses."""



from dataclasses import asdict
from typing import TYPE_CHECKING, Any, Literal

from pydantic.v1 import dataclasses

import datasets
from molflux.modelzoo.info import ModelInfo
from molflux.modelzoo.models.pyod import (
    PyODClassificationMixin,
    PyODModelBase,
    PyODModelConfig,
)
from molflux.modelzoo.typing import PredictionResult
from molflux.modelzoo.utils import get_concatenated_array, pick_features

try:
    if TYPE_CHECKING:
        import numpy as np
    from pyod.models.abod import ABOD
except ImportError as e:
    from molflux.modelzoo.errors import ExtrasDependencyImportError

    raise ExtrasDependencyImportError("pyod", e) from None

ProbabilityConversionMethod = Literal["linear", "unify"]

_DESCRIPTION = """
ABOD class for Angle-base Outlier Detection.

For an observation, the variance of its weighted cosine scores to all
neighbours could be viewed as the outlying score.
See :cite:`kriegel2008angle` for details.

Two version of ABOD are supported:

- Fast ABOD: use k nearest neighbours to approximate.
- Original ABOD: consider all training points with high time complexity at
  O(n^3).
"""

_CONFIG_DESCRIPTION = """
Parameters
----------
contamination : float in (0., 0.5), optional (default=0.1)
    The amount of contamination of the data set, i.e.
    the proportion of outliers in the data set. Used when fitting to
    define the threshold on the decision function.

n_neighbours : int, optional (default=10)
    Number of neighbours to use by default for k neighbours queries.

method: str, optional (default='fast')
    Valid values for metric are:

    - 'fast': fast ABOD. Only consider n_neighbours of training points
    - 'default': original ABOD with all training points, which could be
      slow
"""

Method = Literal["default", "fast"]


class Config:
    arbitrary_types_allowed = True
    extra = "forbid"


@dataclasses.dataclass(config=Config)
class ABODDetectorConfig(PyODModelConfig):
    contamination: float = 0.1
    n_neighbours: int = 10
    method: Method = "fast"


class ABODDetector(PyODClassificationMixin, PyODModelBase[ABODDetectorConfig]):
    @property
    def config(self) -> dict[str, Any]:
        return asdict(self.model_config)

    @property
    def _config_builder(self) -> type[ABODDetectorConfig]:
        return ABODDetectorConfig

    def _info(self) -> ModelInfo:
        return ModelInfo(
            model_description=_DESCRIPTION,
            config_description=_CONFIG_DESCRIPTION,
        )

    def _instantiate_model(self) -> ABOD:
        config = self.model_config
        return ABOD(
            contamination=config.contamination,
            n_neighbors=config.n_neighbours,
            method=config.method,
        )

    def _train(self, train_data: datasets.Dataset, **kwargs: Any) -> Any:
        # Developer note - this function requires special handling as it raises
        # an error on input datasets with non-float columns
        x_data = pick_features(train_data, self.x_features)
        X = get_concatenated_array(x_data, self.x_features)

        # ABOD detectors need float vectors
        X_float: np.ndarray = X.astype(float)

        # instantiate model
        self.model = self._instantiate_model()

        # train (unsupervised learning)
        self.model.fit(X_float)

    def _predict(self, data: datasets.Dataset, **kwargs: Any) -> PredictionResult:
        # Developer note - this function requires special handling as it raises
        # an error on input datasets with non-float columns
        display_names = self._predict_display_names

        if not len(data):
            return {display_name: [] for display_name in display_names}

        x_data = pick_features(data, self.x_features)
        X = get_concatenated_array(x_data, self.x_features)

        # ABOD detectors need float vectors
        X_float: np.ndarray = X.astype(float)
        y_predict: np.ndarray = self.model.predict(X_float, return_confidence=False)

        # Outlier detectors are single task models
        return {display_name: y_predict.tolist() for display_name in display_names}

    def _predict_proba(
        self,
        data: datasets.Dataset,
        method: ProbabilityConversionMethod = "linear",
        **kwargs: Any,
    ) -> PredictionResult:
        # Developer note - this function requires special handling as it raises
        # an error on input datasets with non-float columns

        display_names = self._predict_proba_display_names

        if not len(data):
            return {display_name: [] for display_name in display_names}

        x_data = pick_features(data, self.x_features)
        X = get_concatenated_array(x_data, self.x_features)

        # ABOD detectors need float vectors
        X_float: np.ndarray = X.astype(float)

        y_proba: np.ndarray = self.model.predict_proba(
            X_float,
            method=method,
            return_confidence=False,
        )
        # Outlier detectors are single task models
        return {display_name: y_proba.tolist() for display_name in display_names}



import pytest

from molflux.metrics import Metric, list_metrics, load_metric
from molflux.metrics.regression.spearman import Spearman


@pytest.fixture()
def fixture_metric():
    return load_metric("spearman")


def test_metric_in_catalogue():
    """That the metric is registered in the catalogue."""
    catalogue = list_metrics()
    all_metric_names = [name for names in catalogue.values() for name in names]
    assert "spearman" in all_metric_names


def test_metric_is_mapped_to_correct_class(fixture_metric):
    """That the catalogue name is mapped to the appropriate class."""
    metric = fixture_metric
    assert isinstance(metric, Spearman)


def test_implements_protocol(fixture_metric):
    """That the metric implements the public Metric protocol."""
    metric = fixture_metric
    assert isinstance(metric, Metric)


def test_returns_correlation_and_pvalue(fixture_metric):
    """That that the metric returns two values (correlation and pvalue)."""
    metric = fixture_metric
    predictions = [1, 2, 3, 4, 5]
    references = [5, 6, 7, 8, 7]
    result = metric.compute(predictions=predictions, references=references)
    assert len(result) == 2
    assert "spearman::correlation" in result
    assert "spearman::p_value" in result


def test_default_compute(fixture_metric):
    """That default scoring."""
    metric = fixture_metric
    predictions = [1, 2, 3, 4, 5]
    references = [5, 6, 7, 8, 7]
    result = metric.compute(predictions=predictions, references=references)
    assert pytest.approx(result["spearman::correlation"], 0.01) == 0.82
    assert pytest.approx(result["spearman::p_value"], 0.01) == 0.0886



"""Balanced accuracy metric."""

import logging
from typing import Any

import evaluate
from sklearn.metrics import balanced_accuracy_score

import datasets
from molflux.metrics.bases import HFMetric
from molflux.metrics.typing import ArrayLike, MetricResult

logger = logging.getLogger(__name__)

_DESCRIPTION = """
This metric computes the balanced accuracy, which avoids inflated
performance estimates on imbalanced datasets. It is the macro-average of
recall scores per class or, equivalently, raw accuracy where each sample is
weighted according to the inverse prevalence of its true class. Thus for
balanced datasets, the score is equal to accuracy.
"""

_KWARGS_DESCRIPTION = """
Args:
    predictions: Estimated targets as returned by a classifier.
    references: Ground truth (correct) target values.
    adjusted (optional): When ``True``, the result is adjusted for chance,
            so that random performance would score 0, while keeping perfect
            performance at a score of 1. Defaults to ``False``.
    sample_weight (optional): Weighting of each sample.

Returns:
    balanced_accuracy: balanced accuracy score

Examples:
    >>> from molflux.metrics import load_metric
    >>> metric = load_metric("balanced_accuracy")
    >>> references = [0, 1, 0, 0, 1, 0]
    >>> predictions = [0, 1, 0, 0, 0, 1]
    >>> metric.compute(references=references, predictions=predictions)
    {'balanced_accuracy': 0.625}
"""

_CITATION = """\
@article{scikit-learn,
  title={Scikit-learn: Machine Learning in {P}ython},
  author={Pedregosa, F. and Varoquaux, G. and Gramfort, A. and Michel, V.
         and Thirion, B. and Grisel, O. and Blondel, M. and Prettenhofer, P.
         and Weiss, R. and Dubourg, V. and Vanderplas, J. and Passos, A. and
         Cournapeau, D. and Brucher, M. and Perrot, M. and Duchesnay, E.},
  journal={Journal of Machine Learning Research},
  volume={12},
  pages={2825--2830},
  year={2011}
}
"""


@evaluate.utils.file_utils.add_start_docstrings(_DESCRIPTION, _KWARGS_DESCRIPTION)
class BalancedAccuracy(HFMetric):
    def _info(self) -> evaluate.MetricInfo:
        return evaluate.MetricInfo(
            description=_DESCRIPTION,
            citation=_CITATION,
            inputs_description=_KWARGS_DESCRIPTION,
            features=datasets.Features(
                {
                    "predictions": datasets.Sequence(datasets.Value("int32")),
                    "references": datasets.Sequence(datasets.Value("int32")),
                }
                if self.config_name == "multilabel"
                else {
                    "predictions": datasets.Value("int32"),
                    "references": datasets.Value("int32"),
                },
            ),
            reference_urls=[
                "https://scikit-learn.org/stable/modules/generated/sklearn.metrics.balanced_accuracy_score.html",
            ],
        )

    def _score(
        self,
        *,
        predictions: ArrayLike,
        references: ArrayLike,
        sample_weight: list[float] | None = None,
        adjusted: bool = False,
        **kwargs: Any,
    ) -> MetricResult:
        score = balanced_accuracy_score(
            y_true=references,
            y_pred=predictions,
            adjusted=adjusted,
            sample_weight=sample_weight,
        )
        return {self.tag: float(score)}



from dataclasses import asdict
from typing import Any, Literal

import numpy as np
from pydantic.v1 import dataclasses

from molflux.modelzoo.info import ModelInfo
from molflux.modelzoo.models.pyod import (
    PyODClassificationMixin,
    PyODModelBase,
    PyODModelConfig,
)

try:
    from pyod.models.iforest import IForest
except ImportError as e:
    from molflux.modelzoo.errors import ExtrasDependencyImportError

    raise ExtrasDependencyImportError("pyod", e) from None


_DESCRIPTION = """
Wrapper of scikit-learn Isolation Forest with more functionalities.

The IsolationForest 'isolates' observations by randomly selecting a
feature and then randomly selecting a split value between the maximum and
minimum values of the selected feature.
See :cite:`liu2008isolation,liu2012isolation` for details.

Since recursive partitioning can be represented by a tree structure, the
number of splittings required to isolate a sample is equivalent to the path
length from the root node to the terminating node.

This path length, averaged over a forest of such random trees, is a
measure of normality and our decision function.

Random partitioning produces noticeably shorter paths for anomalies.
Hence, when a forest of random trees collectively produce shorter path
lengths for particular samples, they are highly likely to be anomalies.
"""

_CONFIG_DESCRIPTION = """
Parameters
----------
n_estimators : int, optional (default=100)
    The number of base estimators in the ensemble.

max_samples : int or float, optional (default="auto")
    The number of samples to draw from X to train each base estimator.

        - If int, then draw `max_samples` samples.
        - If float, then draw `max_samples * X.shape[0]` samples.
        - If "auto", then `max_samples=min(256, n_samples)`.

    If max_samples is larger than the number of samples provided,
    all samples will be used for all trees (no sampling).

contamination : float in (0., 0.5), optional (default=0.1)
    The amount of contamination of the data set, i.e. the proportion
    of outliers in the data set. Used when fitting to define the threshold
    on the decision function.

max_features : int or float, optional (default=1.0)
    The number of features to draw from X to train each base estimator.

        - If int, then draw `max_features` features.
        - If float, then draw `max_features * X.shape[1]` features.

bootstrap : bool, optional (default=False)
    If True, individual trees are fit on random subsets of the training
    data sampled with replacement. If False, sampling without replacement
    is performed.

n_jobs : integer, optional (default=1)
    The number of jobs to run in parallel for both `fit` and `predict`.
    If -1, then the number of jobs is set to the number of cores.

random_state : int, random.Generator instance or None, optional (default=None)
    If int, random_state is the seed used by the random number generator;
    If Generator instance, random_state is the random number generator;
    If None, the random number generator is the Generator instance used
    by `np.random`.

verbose : int, optional (default=0)
    Controls the verbosity of the tree building process.
"""

Auto = Literal["auto"]


class Config:
    arbitrary_types_allowed = True
    extra = "forbid"


@dataclasses.dataclass(config=Config)
class IsolationForestDetectorConfig(PyODModelConfig):
    n_estimators: int = 100
    max_samples: float | int | Auto = "auto"
    contamination: float = 0.1
    max_features: float = 1.0
    bootstrap: bool = False
    n_jobs: int = 1
    random_state: int | np.random.Generator | None = None
    verbose: int = 0


class IsolationForestDetector(
    PyODClassificationMixin,
    PyODModelBase[IsolationForestDetectorConfig],
):
    @property
    def config(self) -> dict[str, Any]:
        return asdict(self.model_config)

    @property
    def _config_builder(self) -> type[IsolationForestDetectorConfig]:
        return IsolationForestDetectorConfig

    def _info(self) -> ModelInfo:
        return ModelInfo(
            model_description=_DESCRIPTION,
            config_description=_CONFIG_DESCRIPTION,
        )

    def _instantiate_model(self) -> IForest:
        config = self.model_config
        return IForest(
            n_estimators=config.n_estimators,
            max_samples=config.max_samples,
            contamination=config.contamination,
            max_features=config.max_features,
            bootstrap=config.bootstrap,
            n_jobs=config.n_jobs,
            behaviour="new",
            random_state=config.random_state,
            verbose=config.verbose,
        )



from molflux.features.catalogue import (
    fill_catalogue,
    list_representations,
    register_representation,
)
from molflux.features.load import (
    load_from_dict,
    load_from_dicts,
    load_from_yaml,
    load_representation,
)
from molflux.features.representation import Representation, Representations

# Register all plugins at package import time (to fill catalogue)
fill_catalogue()



from collections.abc import Iterator
from typing import Any, Protocol, runtime_checkable

from molflux.splits.typing import ArrayLike, SplitIndices, Splittable


@runtime_checkable
class SplittingStrategy(Protocol):
    """
    Common API for all splitting strategies.
    """

    def __init__(self, **kwargs: Any) -> None:
        """Initialises the dataset"""

    @property
    def metadata(self) -> dict[str, Any]:
        """Splitting strategy metadata."""

    @property
    def name(self) -> str:
        """The canonical name of the splitting strategy."""

    @property
    def state(self) -> dict[str, Any]:
        """The internal state config."""

    @property
    def tag(self) -> str:
        """The arbitrary tag name of the splitting strategy."""

    def split(
        self,
        dataset: Splittable,
        y: ArrayLike | None = None,
        groups: ArrayLike | None = None,
        **kwargs: Any,
    ) -> Iterator[SplitIndices]:
        """
        Generate indices to split data into training, validation, and test set.

        Args:
            dataset: Training data, of shape (n_samples, n_features).
            y:  The target variable for supervised learning problems. Of shape
                (n_samples, ).
            groups: Group labels for the samples used while splitting the
                dataset into train/test set. Of shape (n_samples, ).

        Yields:
             train_indices: The training set indices for that split.
             validation_indices: The validation set indices for that split.
             test_indices: The validation set indices for that split.
        """

    def reset_state(self) -> None:
        """Resets the state."""

    def update_state(self, **kwargs: Any) -> None:
        """Updates the internal state config."""



import pytest

from molflux.splits.errors import DuplicateKeyError
from molflux.splits.load import (
    load_from_dict,
    load_from_dicts,
    load_from_yaml,
    load_splitting_strategy,
)
from molflux.splits.strategy import SplittingStrategy

representative_strategy_name = "linear_split"


def test_returns_spliting_strategy():
    """That loading a strategy returns an object of type SplittingStrategy."""
    name = representative_strategy_name
    strategy = load_splitting_strategy(name=name)
    assert isinstance(strategy, SplittingStrategy)


def test_forwards_init_kwargs_to_strategy_builder():
    """That keyword arguments get forwarded to the strategy initialiser."""
    name = representative_strategy_name
    strategy = load_splitting_strategy(name=name, tag="pytest-tag")
    assert strategy.tag == "pytest-tag"


def test_invalid_init_kwargs_raise_value_error():
    """That attempting initialisation with invalid keyword arguments raises.

    This is mostly to avoid users passing keyword arguments at initialisation
    time that should have been destined to the .split() method instead. If no
    errors are raised, this could silently lead to bugs as users would expect
    to have preset the splitting strategy with those kwargs.
    """
    name = representative_strategy_name
    with pytest.raises(ValueError, match=r"Unknown initialisation parameter\(s\)"):
        load_splitting_strategy(name=name, invalid_kwarg=True)


def test_from_dict_returns_splitting_strategy():
    """That loading from a dict returns a SplittingStrategy."""
    config = {
        "name": representative_strategy_name,
        "config": {},
        "presets": {},
    }
    strategy = load_from_dict(config)
    assert isinstance(strategy, SplittingStrategy)


def test_from_minimal_dict():
    """That can provide a config with only required fields."""
    config = {
        "name": representative_strategy_name,
    }
    assert load_from_dict(config)


def test_from_dict_forwards_init_kwargs_to_strategy_builder():
    """That config init keyword arguments get forwarded to the strategy initialiser."""
    config = {
        "name": representative_strategy_name,
        "config": {
            "tag": "pytest-tag",
        },
    }
    strategy = load_from_dict(config)
    assert strategy.tag == "pytest-tag"


def test_from_dict_forwards_splitting_kwargs_to_strategy_state():
    """That config splitting keyword arguments get stored in the strategy."""
    config = {
        "name": representative_strategy_name,
        "presets": {
            "train_fraction": 0.123,
        },
    }
    strategy = load_from_dict(config)
    assert strategy.state
    assert strategy.state.get("train_fraction") == 0.123


def test_dict_missing_required_fields_raises():
    """That cannot load a splitting strategy with a config missing required fields."""
    config = {"unknown_key": "value"}
    with pytest.raises(SyntaxError):
        load_from_dict(config)


def test_strategies_with_same_tag_in_collection_raises():
    """That adding several strategies with the same tag in a SplittingStrategies collection raises."""
    config = {
        "name": representative_strategy_name,
    }
    duplicate_configs = [config, config]
    with pytest.raises(DuplicateKeyError):
        load_from_dicts(duplicate_configs)


def test_load_from_dicts_returns_correct_number_of_metrics():
    """That loading from dicts returns a SplittingStrategies collection of the
    expected size."""
    name = representative_strategy_name
    config_one = {
        "name": name,
        "config": {
            "tag": "one",
        },
        "presets": {},
    }
    config_two = {
        "name": name,
        "config": {
            "tag": "two",
        },
        "presets": {},
    }
    configs = [config_one, config_two]
    strategies = load_from_dicts(configs)
    assert len(strategies) == 2


def test_from_yaml_returns_collection_of_splits(fixture_path_to_assets):
    path = fixture_path_to_assets / "config.yml"
    splitting_strategies = load_from_yaml(path=path)
    assert len(splitting_strategies) == 2
    assert "shuffle_split" in splitting_strategies
    assert "custom_k_fold" in splitting_strategies
    assert all(
        isinstance(strategy, SplittingStrategy)
        for strategy in splitting_strategies.values()
    )
    assert splitting_strategies["custom_k_fold"].state.get("n_splits") == 5



from typing import Literal

import numpy.random
from pydantic.v1 import dataclasses

from molflux.modelzoo.info import ModelInfo
from molflux.modelzoo.model import ModelConfig
from molflux.modelzoo.models.sklearn import SKLearnModelBase

try:
    from sklearn.linear_model import Ridge as SKLearnRidge
except ImportError as e:
    from molflux.modelzoo.errors import ExtrasDependencyImportError

    raise ExtrasDependencyImportError("sklearn", e) from e


_DESCRIPTION = """
This is an sklearn linear least squares regression model with l2 regularization.

Minimizes the objective function::

||y - Xw||^2_2 + alpha * ||w||^2_2

This model solves a regression model where the loss function is
the linear least squares function and regularization is given by
the l2-norm. Also known as Ridge Regression or Tikhonov regularization.
This estimator has built-in support for multi-variate regression
(i.e., when y is a 2d-array of shape (n_samples, n_targets)).
"""

_CONFIG_DESCRIPTION = """
Parameters
----------
alpha : {float, ndarray of shape (n_targets,)}, default=1.0
    Constant that multiplies the L2 term, controlling regularization
    strength. `alpha` must be a non-negative float i.e. in `[0, inf)`.
    When `alpha = 0`, the objective is equivalent to ordinary least
    squares, solved by the :class:`LinearRegression` object. For numerical
    reasons, using `alpha = 0` with the `Ridge` object is not advised.
    Instead, you should use the :class:`LinearRegression` object.
    If an array is passed, penalties are assumed to be specific to the
    targets. Hence they must correspond in number.
fit_intercept : bool, default=True
    Whether to fit the intercept for this model. If set
    to false, no intercept will be used in calculations
    (i.e. ``X`` and ``y`` are expected to be centered).
copy_X : bool, default=True
    If True, X will be copied; else, it may be overwritten.
max_iter : int, default=None
    Maximum number of iterations for conjugate gradient solver.
    For 'sparse_cg' and 'lsqr' solvers, the default value is determined
    by scipy.sparse.linalg. For 'sag' solver, the default value is 1000.
    For 'lbfgs' solver, the default value is 15000.
tol : float, default=1e-3
    Precision of the solution.
solver : {'auto', 'svd', 'cholesky', 'lsqr', 'sparse_cg', \
        'sag', 'saga', 'lbfgs'}, default='auto'
    Solver to use in the computational routines:
    - 'auto' chooses the solver automatically based on the type of data.
    - 'svd' uses a Singular Value Decomposition of X to compute the Ridge
      coefficients. It is the most stable solver, in particular more stable
      for singular matrices than 'cholesky' at the cost of being slower.
    - 'cholesky' uses the standard scipy.linalg.solve function to
      obtain a closed-form solution.
    - 'sparse_cg' uses the conjugate gradient solver as found in
      scipy.sparse.linalg.cg. As an iterative algorithm, this solver is
      more appropriate than 'cholesky' for large-scale data
      (possibility to set `tol` and `max_iter`).
    - 'lsqr' uses the dedicated regularized least-squares routine
      scipy.sparse.linalg.lsqr. It is the fastest and uses an iterative
      procedure.
    - 'sag' uses a Stochastic Average Gradient descent, and 'saga' uses
      its improved, unbiased version named SAGA. Both methods also use an
      iterative procedure, and are often faster than other solvers when
      both n_samples and n_features are large. Note that 'sag' and
      'saga' fast convergence is only guaranteed on features with
      approximately the same scale. You can preprocess the data with a
      scaler from sklearn.preprocessing.
    - 'lbfgs' uses L-BFGS-B algorithm implemented in
      `scipy.optimize.minimize`. It can be used only when `positive`
      is True.
    All solvers except 'svd' support both dense and sparse data. However, only
    'lsqr', 'sag', 'sparse_cg', and 'lbfgs' support sparse input when
    `fit_intercept` is True.
positive : bool, default=False
    When set to ``True``, forces the coefficients to be positive.
    Only 'lbfgs' solver is supported in this case.
random_state : int, RandomState instance, default=None
    Used when ``solver`` == 'sag' or 'saga' to shuffle the data.
"""

RidgeRegressorSolver = Literal[
    "auto",
    "svd",
    "cholesky",
    "lsqr",
    "sparse_cg",
    "sag",
    "saga",
    "lbfgs",
]


class Config:
    arbitrary_types_allowed = True
    extra = "forbid"


@dataclasses.dataclass(config=Config)
class RidgeRegressorConfig(ModelConfig):
    alpha: float | list[float] = 1.0
    fit_intercept: bool = True
    copy_X: bool = True
    max_iter: int | None = None
    tol: float = 1e-3
    solver: RidgeRegressorSolver = "auto"
    positive: bool = False
    random_state: None | int | numpy.random.RandomState = None


class RidgeRegressor(SKLearnModelBase[RidgeRegressorConfig]):
    @property
    def _config_builder(self) -> type[RidgeRegressorConfig]:
        return RidgeRegressorConfig

    def _info(self) -> ModelInfo:
        return ModelInfo(
            model_description=_DESCRIPTION,
            config_description=_CONFIG_DESCRIPTION,
        )

    def _instantiate_model(self) -> SKLearnRidge:
        config = self.model_config
        return SKLearnRidge(
            alpha=config.alpha,
            fit_intercept=config.fit_intercept,
            copy_X=config.copy_X,
            max_iter=config.max_iter,
            tol=config.tol,
            solver=config.solver,
            positive=config.positive,
            random_state=config.random_state,
        )



import pytest
from cloudpathlib import implementation_registry
from cloudpathlib.local import LocalS3Client, LocalS3Path, local_s3_implementation


@pytest.fixture
def fixture_test_bucket(monkeypatch):
    """Fixture that patches cloudpathlib to provide a mock S3 bucket.

    References:
        https://cloudpathlib.drivendata.org/stable/testing_mocked_cloudpathlib/
    """

    monkeypatch.setitem(implementation_registry, "s3", local_s3_implementation)

    local_cloud_path = LocalS3Path("s3://cloudpathlib-test-bucket")
    yield local_cloud_path

    # clean up temp directory and replace with new one
    LocalS3Client.reset_default_storage_dir()


@pytest.fixture
def fixture_nested_tmp_path(tmp_path):
    """A nested version of pytest's tmp_path fixture.

    Useful to test operations that should work on non-yet-existent directories.
    """
    yield tmp_path / "sub" / "folder"


@pytest.fixture
def fixture_nested_test_bucket(fixture_test_bucket):
    """A 'nested' test bucket.

    Useful to test operations that should work on non-yet-existent buckets.
    """
    yield fixture_test_bucket / "sub" / "bucket"



from typing import Any

import pandas as pd

import datasets
from molflux.datasets.typing import ExamplesGenerator

_BASE_URL = (
    "https://deepchemdata.s3-us-west-1.amazonaws.com/datasets/delaney-processed.csv"
)

_HOMEPAGE = "https://pubs.acs.org/doi/10.1021/ci034243x"

_DESCRIPTION = """
"""


class ESOL(datasets.GeneratorBasedBuilder):
    def _info(self) -> datasets.DatasetInfo:
        return datasets.DatasetInfo(
            description=_DESCRIPTION,
            features=datasets.Features(
                {
                    "smiles": datasets.Value("string"),
                    "log_solubility": datasets.Value("float64"),
                },
            ),
            homepage=_HOMEPAGE,
        )

    def _split_generators(
        self,
        dl_manager: datasets.DownloadManager,
    ) -> list[datasets.SplitGenerator]:
        archive_path = dl_manager.download(_BASE_URL)

        return [
            datasets.SplitGenerator(
                name="train",
                gen_kwargs={
                    "csv_path": archive_path,
                },
            ),
        ]

    def _generate_examples(
        self,
        csv_path: str | None = None,
        **kwargs: Any,
    ) -> ExamplesGenerator:
        df = pd.read_csv(csv_path)

        # filter unparsable smiles
        df = df[
            (df["smiles"] != "c1c(OC)c(OC)C2C(=O)OCC2c1")
            * (df["smiles"] != "c1c(O)C2C(=O)C3cc(O)ccC3OC2cc1(OC)")
        ]

        for index, row in df.iterrows():
            row_dict = row.to_dict()
            example_dict = {
                "smiles": row_dict["smiles"],
                "log_solubility": row_dict["measured log solubility in mols per litre"],
            }
            yield index, example_dict



import pytest

import datasets
from molflux.datasets.load import load_dataset_builder

representative_builder_name = "esol"


def test_can_load_builder_from_catalogue():
    """That can load a builder from the catalogue."""
    builder = load_dataset_builder(representative_builder_name)
    assert isinstance(builder, datasets.DatasetBuilder)


# TODO: update once we have a multi-config builder to test this better
def test_can_load_specific_builder_config():
    """That can load a specific config of a given builder from the catalogue."""
    builder = load_dataset_builder(representative_builder_name, config_name="default")
    assert isinstance(builder, datasets.DatasetBuilder)


def test_bad_dataset_builder_name_raises_an_error():
    """That a name not in the catalogue raises an error"""
    with pytest.raises(FileNotFoundError):
        load_dataset_builder("molflux_certainly_not_a_dataset")



import warnings
from collections import defaultdict
from collections.abc import Mapping, MutableMapping
from enum import Enum
from typing import Any, Final

import mergedeep
from thefuzz import fuzz

from datasets import Dataset, DatasetDict
from molflux.core.models import get_references, predict
from molflux.core.typing import FoldScores, TasksScores
from molflux.metrics import Metrics
from molflux.modelzoo import Model


def _compute_scores(
    predictions: Dataset,
    references: Dataset,
    metrics: Metrics,
    compute_kwargs: Any,
) -> TasksScores:
    """Computes scores from model predictions and ground truth references.

    Args:
        predictions: The dataset of predictions following the standardised
            predictions format. With predicted features appearing in the same
            order defined by the model architecture.
        references: The corresponding prediction references. Columns are expected
            to match the corresponding order in predictions.
        metrics: The metrics with which to score the model.
        compute_kwargs: A dictionary of extra kwargs to use for computing
            metrics.

    Returns:
        A dictionary with predicted tasks as keys and a dictionary of scores as
        values.
    """

    # some predictions/references might be empty (e.g. from cross-validation strategy splits)
    # return an empty dictionary with empty scores dict for each task
    if not predictions.num_rows or not references.num_rows:
        return dict.fromkeys(references.column_names, {})  # noqa: RUF024

    references_dict, predictions_dict = references.to_dict(), predictions.to_dict()

    tasks_scores: TasksScores = {}
    for (task_name, task_references), (predicted_task_name, task_predictions) in zip(
        references_dict.items(),
        predictions_dict.items(),
        strict=False,
    ):
        # we are allowing predicted tasks to be named differently than
        # reference tasks (e.g. my-model::y1). It's up to the user to make
        # sure that they are ordered correctly to match their respective
        # reference tasks!
        similarity_score = fuzz.token_set_ratio(task_name, predicted_task_name)
        if similarity_score < 100:  # favour false positives
            warnings.warn(
                f"Scores are being computed between reference and predicted tasks with dissimilar names: {task_name!r} vs. {predicted_task_name!r}. Make sure these are equivalent for results to be valid.",
                UserWarning,
                stacklevel=1,
            )

        tasks_scores[task_name] = metrics.compute(
            predictions=task_predictions,
            references=task_references,
            **compute_kwargs,
        )

    return tasks_scores


def compute_scores(
    predictions: DatasetDict,
    references: DatasetDict,
    metrics: Metrics,
    scoring_kwargs: dict[str, dict[str, Any]] | None = None,
) -> FoldScores:
    """Computes scores for a model based on given fold predictions and references.

    Args:
        predictions: The dataset dict of model predictions for each split in the fold.
        references: The dataset dict of model prediction references for each split in the fold.
        metrics: The metrics to score the model with.
        scoring_kwargs: A dictionary of extra kwargs to use for computing
            metrics on each split in the fold.

    Returns:
        A dictionary with split names as keys, and dictionaries of scores for
        each task as values.
    """

    # for each split...
    fold_scores: FoldScores = {}
    for split_name, references_dataset in references.items():
        # make sure malformed kwargs do not go through silently
        if scoring_kwargs and split_name not in scoring_kwargs:
            raise KeyError(f"Please provide scoring kwargs for split {split_name!r}")

        if split_name not in predictions:
            raise KeyError(f"No predictions found for split {split_name!r}")
        predictions_dataset = predictions[split_name]

        kwargs = scoring_kwargs[split_name] if scoring_kwargs else {}
        fold_scores[split_name] = _compute_scores(
            predictions=predictions_dataset,
            references=references_dataset,
            metrics=metrics,
            compute_kwargs=kwargs,
        )

    # Empty splits will have returned empty scores dicts
    # backfill them with metrics names and None scores
    metrics_names = [  # noqa
        list(task_scores.keys())
        for tasks_scores in fold_scores.values()
        for task_scores in tasks_scores.values()
    ][0]
    for split_name, tasks_scores in fold_scores.items():
        for task_name, task_scores in tasks_scores.items():
            if not task_scores:
                fold_scores[split_name][task_name] = dict.fromkeys(metrics_names, None)

    return fold_scores


def score_model(
    model: Model,
    fold: DatasetDict,
    metrics: Metrics,
    prediction_kwargs: dict[str, dict[str, Any]] | None = None,
    scoring_kwargs: dict[str, dict[str, Any]] | None = None,
) -> FoldScores:
    """Scores a model on a given fold (DatasetDict) based on a suite of metrics.

    This is a higher-level wrapper of 'compute_scores', for when accessing or
    storing prediction results is not needed.

    Args:
        model: The model to score.
        fold: The fold to score the model on.
        metrics: The metrics to score the model with.
        prediction_kwargs: A dictionary of extra prediction kwargs to use for
            each split in the fold.
        scoring_kwargs: A dictionary of extra kwargs to use for computing
            metrics on each split in the fold.

    Returns:
        A dictionary with split names as keys, and dictionaries of scores for
        each task as values.
    """

    # other prediction methods would return a tuple of prediction results,
    # which makes scoring ambiguous. Users should use the more fine-grained
    # compute_scores instead for such cases
    prediction_method: Final = "predict"

    references = get_references(model=model, fold=fold)
    predictions = predict(
        model=model,
        fold=fold,
        prediction_method=prediction_method,
        prediction_kwargs=prediction_kwargs,
    )

    return compute_scores(
        predictions=predictions,
        references=references,
        metrics=metrics,
        scoring_kwargs=scoring_kwargs,
    )


def invert_scores_hierarchy(scores: FoldScores) -> FoldScores:
    """Invert a dictionary of scores per task per split into a dictionary
    of scores per split per task, or vice-versa.

    This is a convenience wrapper to be used in situations where it might be
    easier to iterate over scores across tasks rather than splits.
    """
    flipped: defaultdict[str, dict[str, Any]] = defaultdict(dict)
    for key, val in scores.items():
        for subkey, subval in val.items():
            flipped[subkey][key] = subval
    return dict(flipped)


class Strategy(Enum):
    """Merging strategy.

    References:
         https://github.com/clarketm/mergedeep/blob/master/mergedeep/mergedeep.py#L9
    """

    REPLACE = 0
    ADDITIVE = 1
    TYPESAFE = 2
    TYPESAFE_REPLACE = 3
    TYPESAFE_ADDITIVE = 4


def merge_scores(
    destination: MutableMapping,
    *sources: Mapping,
    strategy: Strategy = Strategy.REPLACE,
) -> FoldScores:
    return mergedeep.merge(destination, *sources, strategy=strategy)  # type: ignore[no-any-return]



import pandas as pd
import pytest

import datasets
from molflux.modelzoo.load import load_model

df = pd.DataFrame({"x1": ["Hello"], "y1": ["Buongiorno"]})


@pytest.mark.parametrize(
    "data",
    [
        df,
        datasets.Dataset.from_pandas(df),
    ],
)
def test_predict_from_dataframe(data):
    """That can predict using any dataframe-like object that implements the DataFrame Interchange Protocol."""
    model = load_model("mock_model", x_features=["x1"], y_features=["y1"])
    model.train(data)
    prediction = model.predict(data=data)
    assert prediction == {model.tag: "Spooky Result!"}


def test_invalid_kwarg_raises_value_error():
    """That attempting prediction with invalid keyword arguments raises."""
    model = load_model("mock_model", x_features=["x1"])
    data = datasets.Dataset.from_dict({"x1": ["Hello"], "y1": ["Buongiorno"]})
    model.train(data)
    with pytest.raises(ValueError, match=r"Unknown predict parameter\(s\)"):
        model.predict(data, invalid_kwarg=True)



from __future__ import annotations

import pytest

from molflux.features import Representation, list_representations, load_representation
from molflux.features.representations.rdkit.fingerprints.toxicophores import (
    Toxicophores,
)

representation_name = "toxicophores"


@pytest.fixture(scope="function")
def fixture_representation() -> Representation:
    return load_representation(representation_name)


def test_representation_in_catalogue():
    """That the representation is registered in the catalogue."""
    catalogue = list_representations()
    all_representation_names = [name for names in catalogue.values() for name in names]
    assert representation_name in all_representation_names


def test_representation_is_mapped_to_correct_class(fixture_representation):
    """That the catalogue name is mapped to the appropriate class."""
    representation = fixture_representation
    assert isinstance(representation, Toxicophores)


def test_implements_protocol(fixture_representation):
    """That the representation implements the public Representation protocol."""
    representation = fixture_representation
    assert isinstance(representation, Representation)


def test_default_representation_tag_matches_entrypoint_name(fixture_representation):
    """That the default representation tag matches the entrypoint name.

    This is not strictly required, but ensures a more intuitive user experience.
    """
    representation = fixture_representation
    assert representation.tag == representation_name


def test_compute_one(fixture_representation):
    """That single scoring gives expected results."""
    representation = fixture_representation
    samples = ["C1NC1"]
    result = representation.featurise(samples)
    expected_result = [
        [
            0,
            0,
            0,
            0,
            0,
            0,
            1,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
        ],
    ]
    assert result[representation.tag] == expected_result


def test_compute_batch(fixture_representation):
    """That batch scoring gives expected results."""
    representation = fixture_representation
    samples = ["C1NC1", "CC(=CCCC(=CC=O)C)C"]
    result = representation.featurise(samples)
    expected_result = [
        [
            0,
            0,
            0,
            0,
            0,
            0,
            1,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
        ],
        [
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            1,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
        ],
    ]
    assert result[representation.tag] == expected_result


def test_compute_zero(fixture_representation):
    """That empty scoring gives expected results.

    This should be allowed, and not fail.
    """
    representation = fixture_representation
    samples: list = []
    result = representation.featurise(samples)
    expected_result: list = []
    assert result[representation.tag] == expected_result



"""Tanimoto similarity splitting strategy.

References:
    https://github.com/deepchem/deepchem/blob/master/deepchem/splits/splitters.py#L1193
"""

import logging
from collections.abc import Iterator
from typing import Any

import numpy as np

try:
    from rdkit import Chem, DataStructs
    from rdkit.Chem import AllChem
except ImportError as e:
    from molflux.splits.errors import ExtrasDependencyImportError

    raise ExtrasDependencyImportError("rdkit", e) from e

from molflux.splits.bases import SplittingStrategyBase
from molflux.splits.info import SplittingStrategyInfo
from molflux.splits.typing import ArrayLike, SplitIndices, Splittable

logger = logging.getLogger(__name__)

_DESCRIPTION = """
Class for doing data splits based on the Tanimoto similarity between ECFP4 fingerprints.

This class tries to split the data such that the molecules in each dataset are
as different as possible from the ones in the other datasets.  This makes it a
very stringent test of models.  Predicting the test and validation sets may
require extrapolating far outside the training data.

The running time for this splitter scales as O(n^2) in the number of samples.
Splitting large datasets can take a long time.

Note:
    This strategy requires rdkit to be installed.
"""


class Tanimoto(SplittingStrategyBase):
    def _info(self) -> SplittingStrategyInfo:
        return SplittingStrategyInfo(
            description=_DESCRIPTION,
        )

    def _split(
        self,
        dataset: Splittable,
        y: ArrayLike | None = None,
        groups: ArrayLike | None = None,
        n_splits: int = 1,
        train_fraction: float = 0.8,
        validation_fraction: float = 0.1,
        test_fraction: float = 0.1,
        **kwargs: Any,
    ) -> Iterator[SplitIndices]:
        """Splits compounds according to the Tanimoto similarity of their ECFP4 fingerprints.

        This splitting algorithm has an O(N^2) run time, where N is the number
        of molfluxs in the dataset.

        Args:
            dataset: The data to be split.
            y (optional): List of smiles to be used for tanimoto split.
            groups (optional): Group labels for the samples used while splitting the dataset.
            n_splits (optional): The number of splits to generate. Must be greater than one. Defaults to 2.
            train_fraction: The proportion of the dataset to include in the train split.
            validation_fraction: The proportion of the dataset to include in the validation split.
            test_fraction: The proportion of the dataset to include in the test split.

        Yields:
            A tuple of train, validation, and test indices.

        Examples:
            >>> from molflux.splits import load_splitting_strategy
            >>> strategy = load_splitting_strategy('tanimoto_rdkit')
            >>> dataset = ['CCCC', 'CC', 'c1ccncc1']
            >>> folds = strategy.split(dataset=dataset, y=dataset)
        """

        if y is None:
            raise ValueError("""y parameter should be provided for tanimoto splits.""")

        np.testing.assert_almost_equal(
            train_fraction + validation_fraction + test_fraction,
            1.0,
        )

        try:
            mols = [Chem.MolFromSmiles(smiles) for smiles in y]
        except TypeError as e:
            raise TypeError(
                "Tanimoto splitting strategy expects a collection of SMILES as input.",
            ) from e

        fingerprints = [AllChem.GetMorganFingerprintAsBitVect(x, 2, 1024) for x in mols]

        # Split into two groups: training set and everything else.

        train_size = int(train_fraction * len(dataset))
        validation_size = int(validation_fraction * len(dataset))
        test_size = len(dataset) - train_size - validation_size
        train_indices, test_validation_indices = _split_fingerprints(
            fingerprints,
            train_size,
            validation_size + test_size,
        )

        # Split the second group into validation and test sets.

        if validation_size == 0:
            validation_indices = []
            test_indices = test_validation_indices
        elif test_size == 0:
            test_indices = []
            validation_indices = test_validation_indices
        else:
            test_valid_fps = [fingerprints[i] for i in test_validation_indices]
            test_indices, validation_indices = _split_fingerprints(
                test_valid_fps,
                test_size,
                validation_size,
            )
            test_indices = [test_validation_indices[i] for i in test_indices]
            validation_indices = [
                test_validation_indices[i] for i in validation_indices
            ]

        for _ in range(n_splits):
            yield train_indices, validation_indices, test_indices


def _split_fingerprints(
    fps: list,
    size1: int,
    size2: int,
) -> tuple[list[int], list[int]]:
    """Divides a list of fingerprints into two groups."""

    if len(fps) != size1 + size2:
        raise ValueError(
            f"Incompatible fingerprint splitting sizes for dataset of length {len(fps)}.",
        )

    # Begin by assigning the first molecule to the first group.

    fp_in_group = [[fps[0]], []]
    indices_in_group: tuple[list[int], list[int]] = ([0], [])
    remaining_fp = fps[1:]
    remaining_indices = list(range(1, len(fps)))
    max_similarity_to_group = [
        DataStructs.BulkTanimotoSimilarity(fps[0], remaining_fp),
        [0] * len(remaining_fp),
    ]
    while len(remaining_fp) > 0:
        # Decide which group to assign a molecule to.

        group = 0 if len(fp_in_group[0]) / size1 <= len(fp_in_group[1]) / size2 else 1

        # Identify the unassigned molecule that is least similar to everything in
        # the other group.

        i = np.argmin(max_similarity_to_group[1 - group])

        # Add it to the group.

        fp = remaining_fp[i]
        fp_in_group[group].append(fp)
        indices_in_group[group].append(remaining_indices[i])

        # Update the data on unassigned molecules.

        similarity = DataStructs.BulkTanimotoSimilarity(fp, remaining_fp)
        max_similarity_to_group[group] = np.delete(
            np.maximum(similarity, max_similarity_to_group[group]),
            i,
        )
        max_similarity_to_group[1 - group] = np.delete(
            max_similarity_to_group[1 - group],
            i,
        )
        del remaining_fp[i]
        del remaining_indices[i]

    return indices_in_group



"""
Tests for API features advertised as part of the SplittingStrategy protocol.
"""

from typing import Protocol

from molflux.splits.strategy import SplittingStrategy


def test_splitting_strategy_is_protocol():
    """That SplittingStrategy is a protocol."""
    assert issubclass(type(SplittingStrategy), type(Protocol))


def test_splitting_strategy_has_metadata():
    """That SplittingStrategy has a metadata property."""
    assert hasattr(SplittingStrategy, "metadata")
    assert not callable(SplittingStrategy.metadata)


def test_splitting_strategy_has_name():
    """That SplittingStrategy has a name property."""
    assert hasattr(SplittingStrategy, "name")
    assert not callable(SplittingStrategy.name)


def test_splitting_strategy_has_state():
    """That SplittingStrategy has a state property."""
    assert hasattr(SplittingStrategy, "state")
    assert not callable(SplittingStrategy.state)


def test_splitting_strategy_has_tag():
    """That SplittingStrategy has a tag property."""
    assert hasattr(SplittingStrategy, "tag")
    assert not callable(SplittingStrategy.tag)


def test_splitting_strategy_has_init():
    """That SplittingStrategy has a __init__ method."""
    assert hasattr(SplittingStrategy, "__init__")
    assert callable(SplittingStrategy.__init__)


def test_splitting_strategy_has_split():
    """That SplittingStrategy has a split method."""
    assert hasattr(SplittingStrategy, "split")
    assert callable(SplittingStrategy.split)


def test_splitting_strategy_has_reset_state():
    """That SplittingStrategy has a reset_state method."""
    assert hasattr(SplittingStrategy, "reset_state")
    assert callable(SplittingStrategy.reset_state)


def test_splitting_strategy_has_update_state():
    """That SplittingStrategy has an update_state method."""
    assert hasattr(SplittingStrategy, "update_state")
    assert callable(SplittingStrategy.update_state)



from collections.abc import Iterable, Iterator
from typing import Any

import sklearn.model_selection

from molflux.splits.bases import SplittingStrategyBase
from molflux.splits.info import SplittingStrategyInfo
from molflux.splits.typing import ArrayLike, SplitIndices, Splittable

_DESCRIPTION = """
Leave P Group(s) Out cross-validator

Provides train/validation indices to split data according to a third-party
provided group. This group information can be used to encode arbitrary
domain specific stratifications of the samples as integers.

For instance the groups could be the year of collection of the samples
and thus allow for cross-validation against time-based splits.

The difference between leave_p_groups_out and leave_one_group_out is that
the former builds the validation sets with all the samples assigned to
``p`` different values of the groups while the latter uses samples
all assigned the same groups.
"""


class LeavePGroupsOut(SplittingStrategyBase):
    def _info(self) -> SplittingStrategyInfo:
        return SplittingStrategyInfo(
            description=_DESCRIPTION,
        )

    def _split(
        self,
        dataset: Splittable,
        y: ArrayLike | None = None,
        groups: ArrayLike | None = None,
        p: int | None = None,
        **kwargs: Any,
    ) -> Iterator[SplitIndices]:
        """
        Args:
            dataset: The data to be split.
            groups: Group labels for the samples used while splitting the dataset.
            p: Number of groups (``p``) to leave out in the validation split.

        Examples:
            >>> from molflux.splits import load_splitting_strategy
            >>> strategy = load_splitting_strategy('leave_p_groups_out')
            >>> dataset = range(10)
            >>> groups = [1, 1, 2, 2, 2, 3, 3, 3, 3]
            >>> folds = strategy.split(dataset, groups=groups, p=1)
        """
        if p is None:
            raise ValueError("The 'p' parameter should not be None.")

        if p == 0:
            raise ValueError("The 'p' parameter should not be 0.")

        # Wrap scikit-learn implementation
        cv = sklearn.model_selection.LeavePGroupsOut(n_groups=p)

        for indices in cv.split(dataset, y=y, groups=groups):
            train_indices, validation_indices = indices
            test_indices: Iterable[int] = []  # no test in cross-validation
            yield train_indices, validation_indices, test_indices



from collections.abc import Iterable, Sized
from os import PathLike as OSPathLike
from typing import Union

ArrayLike = Iterable
PathLike = Union[str, OSPathLike]
Splittable = Sized
SplitIndices = tuple[Iterable[int], Iterable[int], Iterable[int]]



import pytest

from molflux.datasets.utils import is_cloud_data


def test_is_cloud_data_none():
    """That None returns false"""
    assert is_cloud_data(None) is False


def test_is_cloud_data_list():
    """That lists are handled appropriately"""
    file_list = ["s3://test-bucket", "/local_dir/", "temp_file.csv"]

    assert is_cloud_data(file_list)
    assert is_cloud_data(file_list[0])
    assert is_cloud_data(file_list[1]) is False
    assert is_cloud_data(file_list[2]) is False


def test_is_cloud_data_bad_input():
    """That dictionaries with non string inputs raise an error"""
    with pytest.raises(
        ValueError,
        match=r".*Could not resolve if data files are cloud data files.*",
    ):
        is_cloud_data({"dictionary_with_non_str_values": 100})  # type: ignore



from molflux.splits.presets import train_test_split


def test_strategy_generates_train_test_splits():
    """That train, and test splits are generated, but no validation."""
    strategy = train_test_split()
    data = range(100)
    folds = strategy.split(data)
    train_indices, validation_indices, test_indices = next(folds)
    assert len(list(train_indices))
    assert len(list(validation_indices)) == 0
    assert len(list(test_indices))


def test_ensure_empty_validation_on_small_samples():
    """That the validation split is empty, even on small samples.

    (Where rounding errors might be more pronounced)
    """
    strategy = train_test_split()

    data = range(4)
    folds = strategy.split(data)
    train_indices, validation_indices, test_indices = next(folds)
    assert len(list(validation_indices)) == 0

    data = range(3)
    folds = strategy.split(data)
    train_indices, validation_indices, test_indices = next(folds)
    assert len(list(validation_indices)) == 0

    data = range(2)
    folds = strategy.split(data)
    train_indices, validation_indices, test_indices = next(folds)
    assert len(list(validation_indices)) == 0



import numpy as np
import pytest

from molflux.splits.catalogue import list_splitting_strategies
from molflux.splits.load import load_splitting_strategy
from molflux.splits.strategy import SplittingStrategy

strategy_name = "leave_one_group_out"


@pytest.fixture(scope="module")
def fixture_test_strategy():
    return load_splitting_strategy(strategy_name)


@pytest.fixture(scope="module")
def fixture_sample_inputs():
    dataset = range(10)
    groups = np.array([1, 1, 2, 2, 2, 3, 3, 3, 3, 3])
    return dataset, groups


def test_is_in_catalogue():
    """That the strategy is registered in the catalogue."""
    catalogue = list_splitting_strategies()
    all_strategy_names = [name for names in catalogue.values() for name in names]
    assert strategy_name in all_strategy_names


def test_implements_protocol(fixture_test_strategy):
    """That the strategy implements the protocol."""
    strategy = fixture_test_strategy
    assert isinstance(strategy, SplittingStrategy)


def test_requires_groups(fixture_test_strategy):
    """That you need to provide groups to use this strategy."""
    dataset = [0, 1, 2, 3, 4, 5]
    strategy = fixture_test_strategy
    with pytest.raises(ValueError):
        next(strategy.split(dataset))


def test_raises_if_only_one_group(fixture_test_strategy):
    """That you need to provide at least two groups to use this strategy."""
    dataset = [0, 1, 2, 3, 4, 5]
    groups = [1, 1, 1, 1, 1, 1]
    strategy = fixture_test_strategy
    with pytest.raises(
        ValueError,
        match="The groups parameter contains fewer than 2 unique groups",
    ):
        next(strategy.split(dataset, groups=groups))


def test_expected_result(fixture_test_strategy):
    """That the splitting strategy yields the expected result in this case.

    Note, we are testing that the splits are not shuffled and that groups are
    held out sequentially in the validation set - starting with the group that
    has the lowest index in the groups array.
    """
    dataset = range(10)
    groups = [2, 2, 2, 3, 3, 3, 3, 3, 1, 1]
    strategy = fixture_test_strategy
    indices = strategy.split(dataset=dataset, groups=groups)

    train_indices, validation_indices, _ = next(indices)
    assert all(
        i == j for i, j in zip(train_indices, [0, 1, 2, 3, 4, 5, 6, 7], strict=False)
    )
    assert all(i == j for i, j in zip(validation_indices, [8, 9], strict=False))

    train_indices, validation_indices, _ = next(indices)
    assert all(
        i == j for i, j in zip(train_indices, [3, 4, 5, 6, 7, 8, 9], strict=False)
    )
    assert all(i == j for i, j in zip(validation_indices, [0, 1, 2], strict=False))

    train_indices, validation_indices, _ = next(indices)
    assert all(i == j for i, j in zip(train_indices, [0, 1, 2, 8, 9], strict=False))
    assert all(
        i == j for i, j in zip(validation_indices, [3, 4, 5, 6, 7], strict=False)
    )


def test_does_not_produce_test_indices(fixture_sample_inputs, fixture_test_strategy):
    """That the splitting strategy is a cross validation strategy."""
    dataset, groups = fixture_sample_inputs
    strategy = fixture_test_strategy
    indices = list(strategy.split(dataset=dataset, groups=groups))
    assert len(indices) > 1


def test_yields_more_than_one_fold(fixture_sample_inputs, fixture_test_strategy):
    """That the splitting strategy yields several folds."""
    dataset, groups = fixture_sample_inputs
    strategy = fixture_test_strategy
    indices = list(strategy.split(dataset=dataset, groups=groups))
    for train_indices, validation_indices, test_indices in indices:
        assert len(train_indices)
        assert len(validation_indices)
        assert not len(test_indices)


def test_yields_correct_number_of_folds(fixture_sample_inputs, fixture_test_strategy):
    """That the splitting strategy yields the correct number of folds.

    There should be as many folds as there are number of unique groups.
    """
    dataset, groups = fixture_sample_inputs
    strategy = fixture_test_strategy
    n_unique_groups = len(np.unique(groups))
    indices = list(strategy.split(dataset=dataset, groups=groups))
    assert len(indices) == n_unique_groups


def test_splits_are_disjoint(fixture_sample_inputs, fixture_test_strategy):
    """That data is spread across splits without overlap."""
    dataset, groups = fixture_sample_inputs
    strategy = fixture_test_strategy
    indices = strategy.split(dataset=dataset, groups=groups)
    for train_indices, validation_indices, _ in indices:
        assert set(train_indices).isdisjoint(set(validation_indices))


def test_each_sample_in_a_different_fold(fixture_sample_inputs, fixture_test_strategy):
    """That each sample ends up in a different fold."""
    dataset, groups = fixture_sample_inputs
    strategy = fixture_test_strategy
    indices = strategy.split(dataset, groups=groups)
    for train_indices, validation_indices, _ in indices:
        groups_in_train = set(groups[train_indices])
        groups_in_validation = set(groups[validation_indices])
        assert set(groups_in_train).isdisjoint(set(groups_in_validation))


def test_deterministic(fixture_sample_inputs, fixture_test_strategy):
    """That repeated splits are deterministic."""
    dataset, groups = fixture_sample_inputs
    strategy = fixture_test_strategy

    # Do one iteration of k-fold splits
    indices_a = strategy.split(dataset, groups=groups)
    train_indices_a1, validation_indices_a1, _ = next(indices_a)
    train_indices_a2, validation_indices_a2, _ = next(indices_a)

    # And again...
    indices_b = strategy.split(dataset, groups=groups)
    train_indices_b1, validation_indices_b1, _ = next(indices_b)
    train_indices_b2, validation_indices_b2, _ = next(indices_b)

    # Check that we get the same results across folds
    assert all(i == j for i, j in zip(train_indices_a1, train_indices_b1, strict=False))
    assert all(
        i == j
        for i, j in zip(validation_indices_a1, validation_indices_b1, strict=False)
    )
    assert all(i == j for i, j in zip(train_indices_a2, train_indices_b2, strict=False))
    assert all(
        i == j
        for i, j in zip(validation_indices_a2, validation_indices_b2, strict=False)
    )


def test_string_groups(fixture_test_strategy):
    """That the splitting strategy yields the expected result when using string
    group labels.

    This would be the use case when using this splitting strategy to e.g. leave
    one 'project' out. We expect the groups to be assigned to the validation
    split in alphabetical order.
    """
    dataset = range(10)
    groups = [
        "P00-00002",
        "P00-00002",
        "P00-00002",
        "P00-00003",
        "P00-00003",
        "P00-00003",
        "P00-00003",
        "P00-00003",
        "P00-00001",
        "P00-00001",
    ]
    strategy = fixture_test_strategy
    indices = strategy.split(dataset=dataset, groups=groups)

    train_indices, validation_indices, _ = next(indices)
    assert all(
        i == j for i, j in zip(train_indices, [0, 1, 2, 3, 4, 5, 6, 7], strict=False)
    )
    assert all(i == j for i, j in zip(validation_indices, [8, 9], strict=False))

    train_indices, validation_indices, _ = next(indices)
    assert all(
        i == j for i, j in zip(train_indices, [3, 4, 5, 6, 7, 8, 9], strict=False)
    )
    assert all(i == j for i, j in zip(validation_indices, [0, 1, 2], strict=False))

    train_indices, validation_indices, _ = next(indices)
    assert all(i == j for i, j in zip(train_indices, [0, 1, 2, 8, 9], strict=False))
    assert all(
        i == j for i, j in zip(validation_indices, [3, 4, 5, 6, 7], strict=False)
    )



from typing import Any

from molflux.features.bases import RepresentationBase
from molflux.features.info import RepresentationInfo
from molflux.features.typing import SmilesArray
from molflux.features.utils import assert_n_positional_args

_DESCRIPTION = """
A generic representation that returns the character count of a string
"""


class CharacterCount(RepresentationBase):
    def _info(self) -> RepresentationInfo:
        return RepresentationInfo(
            description=_DESCRIPTION,
        )

    def _featurise(
        self,
        *columns: SmilesArray,
        without_hs: bool = False,
        **kwargs: Any,
    ) -> dict[str, list[Any]]:
        """Counts the charactes in each string sample.

        Args:
            samples: The data to featurise. Should consist of strings

        Returns:
            The string lengths

        Examples:
            >>> from molflux.features import load_representation
            >>> representation = load_representation('character_count')
            >>> samples = ["hi", "ho", "hum"]
            >>> representation.featurise(samples)
            {'character_count': [2, 2, 3]}
        """
        assert_n_positional_args(*columns, expected_size=1)
        samples = columns[0]

        if without_hs:
            samples = [sample.replace("H", "").replace("h", "") for sample in samples]

        counts = [len(sample) for sample in samples]

        return {f"{self.tag}": counts}



import functools
import logging
import os
from datetime import datetime, timedelta
from typing import Any

import botocore.session
import pyarrow as pa
from boto3.session import Session
from botocore.credentials import RefreshableCredentials
from dateutil.tz import tzlocal
from fsspec.callbacks import _DEFAULT_CALLBACK
from fsspec.implementations.arrow import ArrowFSWrapper
from pyarrow.fs import AwsStandardS3RetryStrategy, S3RetryStrategy

logger = logging.getLogger(__name__)


class S3FileSystem(ArrowFSWrapper):
    """A fsspec compliant S3 FileSystem interface, backed by pyarrow.

    References:
        https://arrow.apache.org/docs/python/filesystems.html#using-arrow-filesystems-with-fsspec
        https://arrow.apache.org/docs/python/generated/pyarrow.fs.S3FileSystem.html
    """

    root_marker = ""
    protocol = "s3"

    def __init__(
        self,
        aws_access_key_id: str | None = None,
        aws_secret_access_key: str | None = None,
        aws_session_token: str | None = None,
        aws_region_name: str | None = None,
        scheme: str | None = None,
        endpoint_override: str | None = None,
        background_writes: bool = True,
        default_metadata: dict | pa.KeyValueMetadata | None = None,
        role_arn: str | None = None,
        session_name: str | None = None,
        external_id: str | None = None,
        load_frequency: int = 900,
        proxy_options: str | dict | None = None,
        allow_bucket_creation: bool = False,
        allow_bucket_deletion: bool = False,
        botocore_session: botocore.session.Session | None = None,
        boto3_session: Session | None = None,
        retry_strategy: S3RetryStrategy | None = None,
        session_ttl: int = 60 * 60,
        refresh_in: int | None = None,
        **kwargs: Any,
    ) -> None:
        if retry_strategy is None:
            retry_strategy = AwsStandardS3RetryStrategy(max_attempts=3)

        self._refresh_in = refresh_in

        # setup boto3 session for properly fetching credentials from AWS
        self._session = get_refreshable_boto3_session(
            aws_access_key_id=aws_access_key_id,
            aws_secret_access_key=aws_secret_access_key,
            aws_session_token=aws_session_token,
            aws_region_name=aws_region_name,
            profile_name=session_name,
            botocore_session=botocore_session,
            boto3_session=boto3_session,
            session_ttl=session_ttl,
        )

        # pyarrow.fs.S3FileSystem initialiser with only aws credentials as free parameters
        self._pyarrow_fs_S3FileSystem = functools.partial(
            pa.fs.S3FileSystem,
            anonymous=False,
            role_arn=role_arn,
            session_name=session_name,
            external_id=external_id,
            load_frequency=load_frequency,
            region=self._session.region_name,
            scheme=scheme,
            endpoint_override=endpoint_override,
            background_writes=background_writes,
            default_metadata=default_metadata,
            proxy_options=proxy_options,
            allow_bucket_creation=allow_bucket_creation,
            allow_bucket_deletion=allow_bucket_deletion,
            retry_strategy=retry_strategy,
        )

        self._fs = self._get_refreshed_filesystem()
        super().__init__(fs=self._fs, **kwargs)

    def _get_refreshed_filesystem(self) -> pa.fs.S3FileSystem:
        """Returns a filesystem with refreshed AWS credentials."""
        logger.debug("Refreshing AWS credentials...")
        credentials = self._session.get_credentials().get_frozen_credentials()
        return self._pyarrow_fs_S3FileSystem(
            access_key=credentials.access_key,
            secret_key=credentials.secret_key,
            session_token=credentials.token,
        )

    @property
    def refresh_needed(self) -> bool:
        """ "Checks if a credentials refresh is needed.

        By default, credentials will be refreshed if they are going to expire
        within the next 15 minutes.

        References:
            https://github.com/boto/botocore/blob/develop/botocore/credentials.py#L438
        """
        botocore_session = self._session._session
        refreshable_credentials = botocore_session._credentials
        return refreshable_credentials.refresh_needed(refresh_in=self._refresh_in)  # type: ignore[no-any-return]

    @property
    def fs(self) -> pa.fs.S3FileSystem:
        """Returns a filesystem with refreshed credentials (if needed).

        This means that every filesystem operation will check if credentials
        are fresh before actuating - as internally self.fs is used to perform
        such operations.
        """
        if self.refresh_needed:
            self._fs = self._get_refreshed_filesystem()
        return self._fs

    @fs.setter
    def fs(self, value: pa.fs.S3FileSystem) -> None:
        self._fs = value

    def get_file(
        self,
        rpath: Any,
        lpath: Any,
        callback: Any = _DEFAULT_CALLBACK,
        outfile: Any = None,
        **kwargs: Any,
    ) -> None:
        """Copy single remote file to local.

        We re-implement this method here to fix a bug when interacting with
        the datasets library and their use of tqdm. The code below is vendored
        from the original base class implementation, with minimal changes.

        References:
            https://github.com/fsspec/filesystem_spec/blob/45a6aec7da1407243f9767c6ab0cff40efee72eb/fsspec/spec.py#L868
            https://arrow.apache.org/docs/python/generated/pyarrow.NativeFile.html#pyarrow.NativeFile.size
        """
        from fsspec.implementations.local import LocalFileSystem
        from fsspec.utils import isfilelike

        if isfilelike(lpath):
            outfile = lpath
        elif self.isdir(rpath):
            os.makedirs(lpath, exist_ok=True)
            return None

        LocalFileSystem(auto_mkdir=True).makedirs(self._parent(lpath), exist_ok=True)

        with self.open(rpath, "rb", **kwargs) as f1:
            if outfile is None:
                outfile = open(lpath, "wb")

            try:
                # for pyarrow files "size" is a callable, not a property
                # so replaced getattr(f1, "size", None) with the following
                size = f1.size() if hasattr(f1, "size") else None
                callback.set_size(size=size)
                data = True
                while data:
                    data = f1.read(self.blocksize)
                    segment_len = outfile.write(data)
                    if segment_len is None:
                        segment_len = len(data)  # type: ignore[arg-type]
                    callback.relative_update(segment_len)
            finally:
                if not isfilelike(lpath):
                    outfile.close()


def get_refreshable_boto3_session(
    aws_access_key_id: str | None = None,
    aws_secret_access_key: str | None = None,
    aws_session_token: str | None = None,
    aws_region_name: str | None = None,
    profile_name: str | None = None,
    botocore_session: botocore.session.Session | None = None,
    boto3_session: Session | None = None,
    session_ttl: int = 60 * 60,
) -> Session:
    """Returns a boto3 session with automatically refreshing credentials.

    Examples:
        >>> session = get_refreshable_boto3_session()
        >>> client = session.client("s3") # we can cache this client object without worrying about expiring credentials
        >>> credentials = session.get_credentials().get_frozen_credentials()  # or we can use it to generate new credentials

    Args:
        session_ttl: The number of seconds to request a session for. Defaults to
            1 hour as that is the maximum session duration allowed by IAM
            profiles in most cases.

    Returns:
        An automatically refreshing boto3 session.

    References:
        https://dev.to/li_chastina/auto-refresh-aws-tokens-using-iam-role-and-boto3-2cjf
        https://stackoverflow.com/a/69226170
    """

    def _get_session_credentials() -> dict[str, Any]:
        """The refresh callable."""
        if boto3_session is not None:
            session = boto3_session
        else:
            session = Session(
                aws_access_key_id=aws_access_key_id,
                aws_secret_access_key=aws_secret_access_key,
                aws_session_token=aws_session_token,
                region_name=aws_region_name,
                profile_name=profile_name,
                botocore_session=botocore_session,
            )

        credentials = {
            **session.get_credentials().get_frozen_credentials()._asdict(),
            "expiry_time": (
                datetime.now(tzlocal()) + timedelta(seconds=session_ttl)
            ).isoformat(),
        }

        logger.debug("AWS credentials expiry time: %s", credentials["expiry_time"])
        return credentials

    # create session with refreshable credentials
    refreshable_credentials = RefreshableCredentials.create_from_metadata(
        metadata=_get_session_credentials(),
        refresh_using=_get_session_credentials,
        method="sts-assume-role",
    )
    refreshable_botocore_session = botocore.session.get_session()
    refreshable_botocore_session._credentials = refreshable_credentials
    refreshable_botocore_session.set_config_variable(
        "region",
        aws_region_name or os.environ.get("AWS_REGION"),
    )
    refreshable_boto3_session = Session(botocore_session=refreshable_botocore_session)
    return refreshable_boto3_session



import pytest

from molflux.modelzoo.errors import NotTrainedError
from molflux.modelzoo.protocols import Estimator
from molflux.modelzoo.store.io import load_from_store, save_to_store


def test_cannot_save_untrained_model_to_s3(
    fixture_mock_model_untrained,
    fixture_empty_bucket,
):
    """That cannot save an untrained model to s3."""
    model = fixture_mock_model_untrained
    bucket_path = fixture_empty_bucket

    with pytest.raises(NotTrainedError):
        save_to_store(bucket_path, model=model)


def test_model_is_saved_to_s3(fixture_mock_model_trained, fixture_empty_bucket):
    """That can save a trained model to s3"""
    model = fixture_mock_model_trained
    bucket_path = fixture_empty_bucket
    save_to_store(bucket_path, model=model)
    assert True


def test_creates_config_json_file(fixture_mock_model_trained, fixture_empty_bucket):
    """That a config.json file is created on save"""
    model = fixture_mock_model_trained
    bucket_path = fixture_empty_bucket
    save_to_store(bucket_path, model=model)
    expected_config_file = bucket_path / "model_config.json"
    assert expected_config_file.exists()
    assert expected_config_file.is_file()


def test_creates_model_artefacts_dir(fixture_mock_model_trained, fixture_empty_bucket):
    """That a model_artefacts/ sub-directory is created on save"""
    model = fixture_mock_model_trained
    bucket_path = fixture_empty_bucket
    save_to_store(bucket_path, model=model)
    expected_artefacts_dir = bucket_path / "model_artefacts"
    assert expected_artefacts_dir.exists()
    assert expected_artefacts_dir.is_dir()


def test_loaded_from_s3(fixture_mock_model_trained, fixture_empty_bucket):
    """That can load a model from s3"""
    model = fixture_mock_model_trained
    bucket_path = fixture_empty_bucket
    save_to_store(bucket_path, model=model)
    loaded_model = load_from_store(bucket_path)
    assert isinstance(loaded_model, Estimator)


def test_loaded_model_is_invariant(fixture_mock_model_trained, fixture_empty_bucket):
    """That a model loaded from store encodes all the original information."""
    model = fixture_mock_model_trained
    bucket_path = fixture_empty_bucket
    save_to_store(bucket_path, model=model)
    loaded_model = load_from_store(bucket_path)
    assert loaded_model.name == model.name
    assert loaded_model.tag == model.tag
    assert loaded_model.config == model.config



import json
from collections.abc import Mapping

from cloudpathlib import AnyPath

from molflux.core.typing import PathLike


def _make_parents_if_needed(path: PathLike) -> None:
    """Creates missing parents of the given path."""
    parent = AnyPath(path).parent  # type: ignore[attr-defined]
    if not parent.is_dir():
        parent.mkdir(parents=True, exist_ok=True)


def save_dict_to_json(mapping: Mapping, path: PathLike, parents: bool = True) -> str:
    """Saves a dictionary to storage as .json

    Args:
        mapping: The dictionary to save.
        path: The path at which to save the dictionary. Can be local or cloud.
        parents: If true, any missing parents of this path are created as needed

    Returns:
        The path to the persisted file.
    """

    if parents is True:
        _make_parents_if_needed(path)

    with AnyPath(path).open("w") as fp:  # type: ignore[attr-defined]
        json.dump(
            mapping,
            fp,
            indent=4,
            sort_keys=True,
        )
    return str(path)



if __name__ == "__main__":
    import os

    import pandas as pd
    from openeye import oechem

    import datasets

    def write_mols(mols, output_file: str) -> None:  # type: ignore
        # dest must be a valid output format
        if not oechem.OEIsWriteable(output_file):
            raise ValueError(
                f"`{output_file}` is not a supported chemical structure format.",
            )

        num_mols = 0
        with oechem.oemolostream(output_file) as oss:
            for mol in mols:
                # written as a copy since OEWriteMolecule can change the object
                oechem.OEWriteMolecule(oss, mol.CreateCopy())
                num_mols += 1

    dl_manager = datasets.DownloadManager()

    archive_path = dl_manager.download_and_extract(
        {
            "mols_file": "http://ogb-data.stanford.edu/data/lsc/pcqm4m-v2-train.sdf.tar.gz",
            "dataset_file": "https://dgl-data.s3-accelerate.amazonaws.com/dataset/OGB-LSC/pcqm4m-v2.zip",
        },
    )
    mols_file = os.path.join(archive_path["mols_file"], "pcqm4m-v2-train.sdf")
    properties_file = os.path.join(
        archive_path["dataset_file"],
        "pcqm4m-v2/raw/data.csv.gz",
    )

    df = pd.read_csv(properties_file, compression="gzip")
    df = df.iloc[:10]

    os.makedirs("pcqm4m-v2/raw/")
    df.to_csv("pcqm4m-v2/raw/data.csv.gz", index=False, compression="gzip")

    mols = oechem.oemolistream(mols_file).GetOEGraphMols()

    mols_10 = [next(mols) for _ in range(10)]

    write_mols(mols_10, "pcqm4m-v2-train.sdf")



import difflib
import functools
import inspect
import logging
from collections import defaultdict
from collections.abc import Callable
from importlib.metadata import EntryPoint, entry_points

from molflux.metrics.metric import Metric
from molflux.metrics.naming import camelcase_to_snakecase

logger = logging.getLogger(__name__)

MetricT = type[Metric]

# Only ever call this once for performance reasons
NAMESPACE = "molflux.metrics.plugins."

# This is where entrypoints will be registered {<name>: <Entrypoint>}
METRICS_CATALOGUE: dict[str, EntryPoint] = {}


@functools.lru_cache
def fill_catalogue() -> None:
    """Fills the catalogue with entrypoints.

    Only ever call this once for performance reasons.
    """
    entrypoints = [
        entrypoint
        for namespace, entrypoints in entry_points().items()
        if namespace.startswith(NAMESPACE)
        for entrypoint in entrypoints
    ]
    for entrypoint in entrypoints:
        put_metric_entrypoint(metric_entrypoint=entrypoint)


def put_metric_entrypoint(metric_entrypoint: EntryPoint) -> None:
    """Puts a metric entrypoint in the catalogue."""
    metric_name = camelcase_to_snakecase(metric_entrypoint.name)

    if metric_name in METRICS_CATALOGUE:
        if metric_entrypoint == METRICS_CATALOGUE[metric_name]:
            pass
        else:
            raise KeyError(f"Duplicate metric: {metric_name!r}")

    METRICS_CATALOGUE[metric_name] = metric_entrypoint


def get_metric_entrypoint(metric_name: str) -> EntryPoint:
    """Returns a specific metric entrypoint from the catalogue.

    The metric is keyed only by name. This allows to decouple from
    eventual reshuffles of entrypoints across groups.
    """
    metric_entrypoint = METRICS_CATALOGUE.get(metric_name)
    if metric_entrypoint is None:
        msg = f"Metric {metric_name!r} is not available."
        similar = difflib.get_close_matches(metric_name, METRICS_CATALOGUE.keys())
        if similar:
            msg += f" You might be looking for one of these: {similar}"
        raise NotImplementedError(msg)
    return metric_entrypoint


def get_metric_cls(metric_name: str) -> MetricT:
    """Loads a specific metric entrypoint from the catalogue."""
    entrypoint = get_metric_entrypoint(metric_name=metric_name)

    # Might raise (e.g. missing dependency extras)
    metric_cls = entrypoint.load()

    if not isinstance(metric_cls, Metric):
        raise NotImplementedError(
            f"Plugin {entrypoint.value} does not implement Metric protocol.",
        )

    return metric_cls  # type: ignore[return-value]


def list_metrics() -> dict[str, list[str]]:
    """List all available metrics in the catalogue.

    The catalogue is returned as a view of metric names keyed by kind.
    """

    view: dict[str, list[str]] = defaultdict(list)

    for metric_name, metric_entrypoint in METRICS_CATALOGUE.items():
        metric_kind = metric_entrypoint.group.split(".")[-1]
        view[metric_kind].append(metric_name)

    return dict(sorted(view.items()))


def register_metric(kind: str, name: str) -> Callable:
    """Registers a custom metric in the entrypoints catalogue.

    Examples:
        .. code-block:: python

           @register_metric(kind="custom", name="my_metric")
           class CustomMetric:
               ...
    """

    def wrapper(metric_cls: MetricT) -> MetricT:
        module = inspect.getmodule(metric_cls).__name__  # type: ignore
        value = f"{module}:{metric_cls.__name__}"
        group = NAMESPACE + kind
        entrypoint = EntryPoint(name=name, value=value, group=group)
        put_metric_entrypoint(metric_entrypoint=entrypoint)
        return metric_cls

    return wrapper



"""
This file implements the functions mapping json-serializable configs of SKLearn classes to actual python instances.
"""

from importlib import import_module
from importlib.util import find_spec

# https://docs.pydantic.dev/latest/usage/types/#typeddict
from typing import Any, TypedDict

try:
    from sklearn.pipeline import Pipeline, make_pipeline

except ImportError as e:
    from molflux.modelzoo.errors import ExtrasDependencyImportError

    raise ExtrasDependencyImportError("sklearn", e) from e


def _load_class(qualified_class: str) -> Any:
    """
    Returns a class from a string definition of a class.
    Args:
        qualified_class (str): String representation of the class path. For example,
        `sklearn.ensemble.RandomForest`.

    Returns:
        Class of the specified string definition.

    """
    *module_name_comps, class_name = qualified_class.split(".")

    # assert the class has a prefix module
    assert module_name_comps

    module_name = ".".join(module_name_comps)
    if find_spec(module_name) is None:
        raise ValueError(f"failed to find module - {module_name}")

    module = import_module(module_name)
    if not hasattr(module, class_name):
        raise ValueError(
            f"failed to find class = {class_name} in module = {module_name}",
        )

    ret = getattr(module, class_name)

    return ret


class StepConfig(TypedDict):
    """Type definition of the dictionary structure for a single pipeline step in an SKLearnPipeline config"""

    class_name: str
    hyperparameters: dict[str, Any] | None


# type definition of the config expected by SKLearnPipeline models
StepConfigsT = list[StepConfig]


def _get_step_object(step_config: StepConfig) -> Any:
    """
    Return an instance of a class from a given config.

    Args:
        step_config (StepConfig): Dictionary with 2 keys, "class_name" and "hyperparameters".
        This should follow the format:
        {
            "class_name": "sklearn.some_module.ClassName",
            "hyperparameters": {
                "first_hp_for_ClassName": 500,
                "second_hp_for_ClassName": "auto",
                "learning_rate": 0.1,
                ...
            },
        }

    Returns:
        An object equivalent to ClassName(**hyperparameters)

    """
    try:
        # get the class from the specified import path
        class_name = step_config.get("class_name")
        if class_name is None:
            raise ValueError("class_name parameter must be specified")
        model_class = _load_class(class_name)

        # get the hyperparameters for the class
        hyperparams = step_config.get("hyperparameters", {})
        if hyperparams is None:
            hyperparams = {}

        # instantiate an object of the class with the hyperparameters
        return model_class(**hyperparams)
    except Exception as e:
        raise RuntimeError(f"Could not generate step for {step_config=}") from e


def build_pipeline(step_configs: StepConfigsT) -> Pipeline:
    """
    Construct an SKLearn pipeline from a list of step configs.

    Args:
        step_configs (StepConfigsT): List of steps to be formatted into an SKLearn Pipeline.
        This should follow the format:
            [
                {
                    "class_name": "sklearn.some_module.ClassName",
                    "hyperparameters": {
                        "first_hp_for_ClassName": 500,
                        "second_hp_for_ClassName": "auto",
                        "learning_rate": 0.1,
                        ...
                    },
                },
                {
                    "class_name": ...
                    "hyperparameters": ...
                },
                ...
            ]

    Returns: Pipeline
    """
    step_objects = []

    # for each step in the pipeline config, instantiate the corresponding object
    for step_config in step_configs:
        step_object = _get_step_object(step_config)
        step_objects.append(step_object)

    # make the pipeline from each step
    return make_pipeline(*step_objects)



import pytest

from molflux.features import Representation, list_representations, load_representation
from molflux.features.representations.rdkit.fingerprints.topological_torsion import (
    TopologicalTorsion,
)

representation_name = "topological_torsion"


@pytest.fixture(scope="function")
def fixture_representation() -> Representation:
    return load_representation(representation_name)


def test_representation_in_catalogue():
    """That the representation is registered in the catalogue."""
    catalogue = list_representations()
    all_representation_names = [name for names in catalogue.values() for name in names]
    assert representation_name in all_representation_names


def test_representation_is_mapped_to_correct_class(fixture_representation):
    """That the catalogue name is mapped to the appropriate class."""
    representation = fixture_representation
    assert isinstance(representation, TopologicalTorsion)


def test_implements_protocol(fixture_representation):
    """That the representation implements the public Representation protocol."""
    representation = fixture_representation
    assert isinstance(representation, Representation)



import pytest

import datasets
from molflux.datasets.load import load_from_dict

representative_dataset_name = "esol"


def test_returns_dataset():
    """That loading from a dict returns a Dataset."""
    name = representative_dataset_name
    config = {
        "name": name,
        "config": {},
    }
    dataset = load_from_dict(config)
    assert isinstance(dataset, datasets.Dataset)


def test_none_split_returns_dataset_dict():
    """That loading from a dict can also return a DatasetDict if split None."""
    name = representative_dataset_name
    config = {
        "name": name,
        "config": {
            "split": None,
        },
    }
    dataset = load_from_dict(config)
    assert isinstance(dataset, datasets.DatasetDict)


def test_from_minimal_dict():
    """That can provide a config with only required fields."""
    name = representative_dataset_name
    config = {
        "name": name,
    }
    assert load_from_dict(config)


def test_dict_missing_required_fields_raises():
    """That cannot load a dataset with a config missing required fields."""
    config = {"unknown_key": "value"}
    with pytest.raises(SyntaxError):
        load_from_dict(config)



from typing import Any

try:
    from openeye.oemolprop import OEGetRotatableBondCount
except ImportError as e:
    from molflux.features.errors import ExtrasDependencyImportError

    raise ExtrasDependencyImportError("openeye", e) from None

from molflux.features.bases import RepresentationBase
from molflux.features.info import RepresentationInfo
from molflux.features.representations.openeye._utils import to_digit, to_oemol
from molflux.features.typing import MolArray
from molflux.features.utils import assert_n_positional_args, featurisation_error_harness

_DESCRIPTION = """
The number of rotatable bond counts in a molecule.

Rotatable bond is defined as any single non-ring bond, bounded to nonterminal
heavy (i.e. non-hydrogen) atom. In addition, It considers the carbon-carbon
triple bond in acetylene rotatable. The representation excludes single bonds
observed in a certain groups, such as sulfonamides, esters, and amidine.
"""


class RotatableBonds(RepresentationBase):
    def _info(self) -> RepresentationInfo:
        return RepresentationInfo(
            description=_DESCRIPTION,
        )

    def _featurise(
        self,
        *columns: MolArray,
        digitise: bool = False,
        start: int = 0,
        stop: int = 16,
        num: int = 8,
        adjust: bool = False,
        **kwargs: Any,
    ) -> dict[str, list[int]]:
        """Calculates the number of rotatable bond counts for each input molecule.

        Args:
            samples: The sample molecules to featurise.
            digitise: Whether to digitise the results. Defaults to `False`.
            start: If `digitise=True`, the start of the binning range. Defaults
                to 0.
            stop: If `digitise=True`, the end of the binning range. Defaults
                to 16.
            num: If `digitise=True`, the number of bins. Defaults to 8.
            adjust: This parameter allows optional adjustment for aliphatic
                rings following the method of [Oprea-2000]_. In this case, it
                also estimates the number of flexible bonds in aliphatics rings

        Returns:
            The number of rotatable bond counts of each input molecule.

        References:
            [Oprea-2000]_ Oprea, T., Property Distribution of Drug-Related Chemical Databases, Journal of Computer-Aided Molecular Design, Vol. 14, pp. 251-264, 2000

        Examples:
            >>> from molflux.features import load_representation
            >>> representation = load_representation('rotatable_bonds')
            >>> samples = ['C1=CC=C(C=C1)S(=O)O', 'NH']
            >>> representation.featurise(samples=samples)
            {'rotatable_bonds': [1, 0]}
        """
        assert_n_positional_args(*columns, expected_size=1)
        samples = columns[0]
        results = []
        for sample in samples:
            with featurisation_error_harness(sample):
                mol = to_oemol(sample)
                result = OEGetRotatableBondCount(mol=mol, adjust=adjust)

                if digitise:
                    result = to_digit(result, start=start, stop=stop, num=num).tolist()

                results.append(result)

        return {self.tag: results}



import json
import logging
import shutil
import tempfile
import warnings
from pathlib import Path
from typing import Any, Union

from cloudpathlib import AnyPath, CloudPath

from molflux.modelzoo import config
from molflux.modelzoo.load import load_model
from molflux.modelzoo.protocols import Model
from molflux.modelzoo.store.artefact import ModelArtefact

logger = logging.getLogger(__name__)

_AnyPath = Union[Path, CloudPath]


def _rmtree(path: _AnyPath) -> None:
    if isinstance(path, Path):
        shutil.rmtree(str(path))
    elif isinstance(path, CloudPath):
        path.rmtree()
    else:
        raise ValueError(f"Unsupported path type: {type(path)!r}")


def _save_model_metadata(artefact: ModelArtefact, directory: _AnyPath) -> str:
    """Saves model metadata as json to a directory."""
    from molflux import __version__

    metadata = {
        "name": artefact.name,
        "tag": artefact.tag,
        "config": artefact.config,
        "version": __version__,
    }

    log_file = directory / config.MODEL_CONFIG_FILENAME

    if log_file.exists():
        logger.warning("Overwriting existing model metadata: %s", log_file)
    else:
        log_file.parent.mkdir(parents=True, exist_ok=True)

    with log_file.open("w", encoding="utf-8") as f:
        json.dump(metadata, f, indent=4, sort_keys=True)

    return str(log_file)


def _save_binaries(artefact: ModelArtefact, directory: _AnyPath) -> str:
    """Serialises a model into a given directory."""
    model = artefact.model

    # The 'official' subdirectory into which to serialise the model
    artefacts_dir = directory / config.MODEL_ARTEFACT_DIR_FILENAME

    if artefacts_dir.exists():
        logger.warning("Overwriting existing model artefacts: %s", artefacts_dir)
    else:
        artefacts_dir.mkdir(parents=True, exist_ok=True)

    try:
        # if saving to cloud, we will need to upload a temporary directory as
        # models only know how to save themselves to disk
        if isinstance(artefacts_dir, CloudPath):
            with tempfile.TemporaryDirectory() as tmpdir:
                model.as_dir(directory=str(tmpdir))
                ok = artefacts_dir.upload_from(tmpdir, force_overwrite_to_cloud=True)
            if not ok:
                raise RuntimeError("Error uploading model artefact to store.")
        else:
            model.as_dir(directory=str(artefacts_dir))

    except Exception:
        # cleanup and raise exception up the stack
        _rmtree(artefacts_dir)
        raise

    return str(artefacts_dir)


def _load_model_metadata(directory: _AnyPath) -> tuple[str, str, dict[str, Any]]:
    """Loads model metadata from a directory.

    The directory is expected to contain a metadata file as generated
    by `_save_model_metadata()`.
    """
    from molflux import __version__

    log_file = directory / config.MODEL_CONFIG_FILENAME
    with log_file.open("r", encoding="utf-8") as f:
        metadata: dict[str, Any] = json.load(f)

    # Peek at stored model versioning
    version = metadata.get("version")
    if version and version != __version__:
        warnings.warn(
            f"Loading model built by molflux.modelzoo=={version}. "
            f"Behaviour may be different in this version.",
            stacklevel=2,
        )

    # Extract model identifier that will be used to load relevant class
    name = metadata.pop("name", None)
    if name is None:
        raise KeyError("Invalid model metadata: missing `name`.")

    tag = metadata.pop("tag", None)  # not essential

    model_config = metadata.pop("config", None)
    if model_config is None:
        raise KeyError("Invalid model metadata: missing `config`.")

    return name, tag, model_config


def _load_from_binaries(
    directory: _AnyPath,
    name: str,
    tag: str,
    model_config: dict[str, Any],
) -> Model:
    """Deserialises a model that has been serialised in a given directory.

    The binaries are expected to have been created with `_save_binaries()`.
    """

    # The 'official' directory for model binaries
    artefacts_dir = directory / config.MODEL_ARTEFACT_DIR_FILENAME

    # Load model class and then use it to deserialise binaries
    model = load_model(name=name, tag=tag, **model_config)

    # if loading from cloud, we will need to download to a temporary directory
    # as models only know how to load themselves from disk
    if isinstance(artefacts_dir, CloudPath):
        with tempfile.TemporaryDirectory() as tmpdir:
            ok = artefacts_dir.download_to(destination=tmpdir)
            if not ok:
                raise RuntimeError("Error downloading model artefact from store.")
            model.from_dir(directory=str(tmpdir))
    else:
        model.from_dir(directory=str(artefacts_dir))

    return model


def save_artefact_to_store(artefact: ModelArtefact, directory: str) -> None:
    directory_path: _AnyPath = AnyPath(directory)  # type: ignore[assignment]

    # save model config
    _save_model_metadata(artefact=artefact, directory=directory_path)

    # save model binaries
    _save_binaries(artefact=artefact, directory=directory_path)


def load_artefact_from_store(directory: str) -> ModelArtefact:
    directory_path: _AnyPath = AnyPath(directory)  # type: ignore[assignment]
    if not directory_path.exists():
        raise FileNotFoundError(f"No such directory: {directory!r}")
    if not directory_path.is_dir():
        raise ValueError(f"Is not a directory: {directory!r}")

    # load model metadata
    name, tag, model_config = _load_model_metadata(directory=directory_path)

    # load model
    model = _load_from_binaries(
        name=name,
        tag=tag,
        model_config=model_config,
        directory=directory_path,
    )

    return ModelArtefact(
        name=name,
        tag=tag,
        config=model_config,
        model=model,
    )
